from typing import Callable, Sequence

from adaptive_harmony import (
    JobNotifier,
    Logger,
    StageNotifier,
    StringThread,
    TrainingModel,
)
from adaptive_harmony.common.callbacks import RecipeCallback
from adaptive_harmony.common.grpo import GRPO, Sample
from adaptive_harmony.graders import BaseGrader
from adaptive_harmony.metric_logger import StdoutLogger


class GSPO(GRPO):  # grpo already logs args so we don't do it here
    def __init__(
        self,
        dataset: list[StringThread],
        model: TrainingModel,
        grader: BaseGrader,
        logger: Logger = StdoutLogger(),
        stage_notifier: StageNotifier = JobNotifier().stage_notifier("GSPO Training"),
        callbacks: Sequence[RecipeCallback] = [],
        max_num_gspo_steps: int | None = None,
        completions_per_sample=8,
        lr: float = 7.5e-7,
        lr_scheduler: Callable[[float], float] | None = None,
        samples_per_batch=128,
        samples_per_mini_batch=128,
        mini_epochs_per_batch=1,
        max_grad_norm=1.0,
        # wildly different defaults than GRPO because we are looking at the
        # entire sequence at once, I tried the number from the GSPO paper but
        # it was skipping ~60% of the samples. I had good success with 0.01 but
        # it's not properly swept yet.
        clip_range=0.01,
        kl_beta=0.01,
        weight_decay=0.0,
    ):
        super().__init__(
            dataset,
            model,
            grader,
            logger,
            stage_notifier,
            callbacks,
            max_num_grpo_steps=max_num_gspo_steps,
            completions_per_sample=completions_per_sample,
            lr=lr,
            lr_scheduler=lr_scheduler,
            samples_per_batch=samples_per_batch,
            samples_per_mini_batch=samples_per_mini_batch,
            mini_epochs_per_batch=mini_epochs_per_batch,
            max_grad_norm=max_grad_norm,
            clip_range=clip_range,
            kl_beta=kl_beta,
            weight_decay=weight_decay,
        )

    async def train_sample(self, sample: Sample):
        await self.model.train_gspo(
            sample.sample,
            sample.logprobs,
            sample.ref_logprobs,
            [sample.advantage],  # only diff with grpo is we train with a single advantage
            self.clip_range,
            self.clip_range,
            self.kl_beta,
        )



class IgnoreScoreException(Exception):
    """
    Exception to indicate that a score should be ignored/is not applicable.
    Accepts a custom message explaining why the score was ignored.
    """

    def __init__(self, message: str = "Score should be ignored"):
        super().__init__(message)
        self.message = message



from asyncio import gather
from typing import Any, Literal, Sequence

from loguru import logger

from adaptive_harmony import StringThread
from adaptive_harmony.core.structured_output import JsonParseError
from adaptive_harmony.graders import BaseGrader, Grade
from adaptive_harmony.graders.exceptions import IgnoreScoreException


class CombinedGrader(BaseGrader):
    """
    Combines grades from multiple graders.
    Aggregates their results using weighted sum or average.
    Ignores failing graders and proceeds calculating the aggregate grade with the rest.
    """

    def __init__(
        self,
        grader_key: str,
        graders: Sequence[BaseGrader],
        weights: list[float] | None = None,
        aggregation_method: Literal["sum", "mean"] = "mean",
        failure_rate_warning_threshold: float = 0.2,
    ):
        super().__init__(grader_key)
        self.graders = graders
        if weights:
            assert len(weights) == len(graders), "Number of weights must match number of graders"
        self.weights = weights or [1.0] * len(graders)
        self.agg_method = aggregation_method
        self.failure_rate_warning_threshold = failure_rate_warning_threshold

    async def grade(self, sample: StringThread) -> Grade:
        async def separate_success_from_fail_graders(grader: BaseGrader) -> Grade | None:
            try:
                return await grader.grade(sample)
            except (IgnoreScoreException, JsonParseError):
                # return None if score is supposed to be ignored, or judge output format failure
                return None
            except Exception as e:
                # fail for any other exception
                raise e

        tasks = [separate_success_from_fail_graders(grader) for grader in self.graders]
        results: list[Grade | None] = await gather(*tasks)

        weighted_scores = []
        failed_graders = []

        # Separate successful and failed results
        successful_graders: list[str] = []
        successful_grades: list[Grade] = []
        successful_weights = []

        for result, weight, grader in zip(results, self.weights, self.graders):
            if result is not None:
                # Successful grader
                weighted_score = result.value * weight
                successful_graders.append(grader.grader_key)
                weighted_scores.append(weighted_score)
                successful_grades.append(result)
                successful_weights.append(weight)
            else:
                # Failed grader
                failed_graders.append(grader.grader_key)

        # Fail if no successfull graders
        if not successful_grades:
            raise RuntimeError("All graders failed - cannot compute aggregate grade")

        # Warn if more than a set % of scorers failed
        total_graders = len(self.graders)
        failure_rate = len(failed_graders) / total_graders
        if failure_rate > self.failure_rate_warning_threshold:
            logger.warning(f"{len(failed_graders)}/{total_graders}% of graders failed for sample: {failed_graders}")

        # Aggregate scores
        if self.agg_method == "sum":
            final_score = sum(weighted_scores)
        elif self.agg_method == "mean":
            # For average, we normalize by the sum of successful weights (renormalize)
            final_score = sum(weighted_scores) / sum(successful_weights)
        else:
            raise ValueError(f"Unknown aggregation method: {self.agg_method}")

        # Log the combined score.
        self.add_log({"score": final_score})

        reason = "\n".join(
            [
                f"{grader_key}: {grade.value} - {grade.reasoning}"
                for grader_key, grade in zip(successful_graders, successful_grades)
            ]
        )

        return Grade(
            value=final_score,
            grader_key=self.grader_key,
            reasoning=reason,
        )

    def get_logs(self, clear: bool = False, log_all_samples: bool = False) -> dict[str, Any]:
        combined_logs = super().get_logs(clear=False, log_all_samples=log_all_samples)
        all_logs = combined_logs
        for grader in self.graders:
            scorer_logs = grader.get_logs(clear=clear, log_all_samples=log_all_samples)
            all_logs = all_logs | {f"{grader.grader_key}/{key}": value for key, value in scorer_logs.items()}

        if clear:
            self.clear_logs()
        return all_logs

    def clear_logs(self) -> None:
        super().clear_logs()
        for grader in self.graders:
            grader.clear_logs()



import asyncio
import pytest
import json
import tempfile
import os
from adaptive_harmony import StringThread
from adaptive_harmony.parameters import Dataset, dataset_kinds
from adaptive_harmony.runtime.context import RecipeContext
from unittest.mock import AsyncMock, Mock


@pytest.fixture
def mock_ctx():
    """Create a mock RecipeContext for testing."""
    ctx = Mock(spec=RecipeContext)
    ctx.file_storage = Mock()
    ctx.file_storage.mk_url = Mock(return_value="file://test_file.jsonl")

    # Create a mock for dataset config response
    mock_dataset_config = Mock()
    mock_dataset_config.file_path = None

    ctx.client = Mock()
    ctx.client.get_dataset_config = AsyncMock(return_value=mock_dataset_config)
    return ctx


def create_temp_dataset_file(data, mock_ctx):
    """Helper function to create a temporary dataset file and configure the mock."""
    if isinstance(data, list):
        content = "\n".join(json.dumps(item) for item in data) + "\n"
    else:
        content = json.dumps(data) + "\n"

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".jsonl") as f:
        f.write(content)
        temp_file = f.name

    # Configure mock to return file content
    mock_ctx.file_storage.read.return_value = content.encode("utf-8")

    return temp_file


class TestAdaptiveDatasetLoadDataset:

    def test_load_dataset_prompt_format(self, mock_ctx):
        prompt_data = {
            "prompt": [
                ["user", "What is the capital of France?"],
                ["assistant", "The capital of France is Paris."],
            ],
            "metadata": {"id": "b2761a88-0662-453b-aba7-27846434dc55", "created_at": 1234567890},
        }

        temp_file = create_temp_dataset_file(prompt_data, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Prompt](dataset_key="test_prompt", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 1
            assert isinstance(threads[0], StringThread)
            assert str(threads[0].metadata["id"]) == "b2761a88-0662-453b-aba7-27846434dc55"
        finally:
            os.unlink(temp_file)

    def test_load_dataset_completion_format(self, mock_ctx):
        completion_data = {
            "prompt": [["user", "Write a haiku about coding."]],
            "completion": ["assistant", "Code flows like water,\nBugs hide in silent corners,\nDebug light reveals."],
            "metadata": {"id": "b2761a88-0662-453b-aba7-27846434dc56", "created_at": 1234567890},
        }

        temp_file = create_temp_dataset_file(completion_data, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Prompt](dataset_key="test_completion", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 1
            assert isinstance(threads[0], StringThread)
            assert str(threads[0].metadata["id"]) == "b2761a88-0662-453b-aba7-27846434dc56"
            assert "Code flows like water" in threads[0].last_content()
        finally:
            os.unlink(temp_file)

    def test_load_dataset_metric_format(self, mock_ctx):
        metric_data = {
            "prompt": [["user", "Rate this response quality."]],
            "completion": ["assistant", "This is a high-quality response."],
            "metrics": {"quality_score": 0.95, "helpfulness": 0.88},
            "metadata": {
                "id": "b2761a88-0662-453b-aba7-27846434dc57",
                "created_at": 1234567890,
                "annotator": "expert_1",
            },
        }

        temp_file = create_temp_dataset_file(metric_data, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Metric](
                dataset_key="test_metric", local_file_path=temp_file, feedback_key="quality_score"
            )
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 1
            assert isinstance(threads[0], StringThread)
            assert str(threads[0].metadata["id"]) == "b2761a88-0662-453b-aba7-27846434dc57"
            assert threads[0].metadata["res"] == 0.95
        finally:
            os.unlink(temp_file)

    def test_load_dataset_preference_format(self, mock_ctx):
        preference_data = {
            "prompt": [["user", "Explain quantum computing."]],
            "good_completion": ["assistant", "Quantum computing uses quantum mechanical phenomena..."],
            "bad_completion": ["assistant", "I don't know."],
            "metadata": {
                "id": "b2761a88-0662-453b-aba7-27846434dc58",
                "created_at": 1234567890,
                "preference_strength": "strong",
            },
        }

        temp_file = create_temp_dataset_file(preference_data, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Preference](dataset_key="test_preference", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 1
            assert isinstance(threads[0], StringThread)
            assert str(threads[0].metadata["id"]) == "b2761a88-0662-453b-aba7-27846434dc58"
            assert "Quantum computing uses" in threads[0].metadata["preferred_completion"]
            assert "I don't know" in threads[0].metadata["other_completion"]
        finally:
            os.unlink(temp_file)

    def test_load_dataset_external_format(self, mock_ctx):
        external_data = {
            "messages": [
                {"role": "user", "content": "Hello, how are you?"},
                {"role": "assistant", "content": "I'm doing well, thank you!"},
            ],
            "completion": "What a nice day",
            "metadata": {"source": "external_dataset", "quality": "high"},
        }

        temp_file = create_temp_dataset_file(external_data, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Prompt](dataset_key="test_external", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 1
            assert isinstance(threads[0], StringThread)
            assert threads[0].metadata["source"] == "external_dataset"
            assert threads[0].metadata["quality"] == "high"
            assert "What a nice day" in threads[0].last_content()
        finally:
            os.unlink(temp_file)

    def test_load_dataset_external_format_with_preferences(self, mock_ctx):
        """Test loading dataset with external format including preference data."""
        external_data = {
            "messages": [{"role": "user", "content": "Tell me about AI."}],
            "other_completion": "AI is bad.",
            "preferred_completion": "AI is a powerful technology that can help solve complex problems.",
            "metadata": {"topic": "AI_discussion"},
        }

        temp_file = create_temp_dataset_file(external_data, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Preference](dataset_key="test_external_pref", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 1
            assert isinstance(threads[0], StringThread)
            assert threads[0].metadata["topic"] == "AI_discussion"
            assert threads[0].metadata["other_completion"] == "AI is bad."
            assert (
                threads[0].metadata["preferred_completion"]
                == "AI is a powerful technology that can help solve complex problems."
            )
        finally:
            os.unlink(temp_file)

    def test_load_dataset_multiple_lines(self, mock_ctx):
        """Test loading dataset with multiple lines."""
        data_lines = [
            {"input": [{"role": "user", "content": "Question 1?"}], "completion": "Answer 1", "metadata": {"id": 1}},
            {"input": [{"role": "user", "content": "Question 2?"}], "completion": "Answer 2", "metadata": {"id": 2}},
        ]

        temp_file = create_temp_dataset_file(data_lines, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Completion](dataset_key="test_multiple", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 2
            assert all(isinstance(thread, StringThread) for thread in threads)
            assert threads[0].metadata["id"] == 1
            assert threads[1].metadata["id"] == 2
            assert "Answer 1" in threads[0].last_content()
            assert "Answer 2" in threads[1].last_content()
        finally:
            os.unlink(temp_file)

    def test_load_dataset_no_valid_samples_raises_error(self, mock_ctx):
        """Test that loading a dataset with no valid samples raises ValueError."""
        empty_content = "\n   \n"  # Only empty lines
        mock_ctx.file_storage.read.return_value = empty_content.encode("utf-8")

        dataset = Dataset[dataset_kinds.Prompt](dataset_key="test_empty", local_file_path="dummy_file")

        with pytest.raises(ValueError, match="Did not find any valid format samples in the dataset"):
            asyncio.run(dataset.load(mock_ctx))

    def test_load_dataset_external_metadata_with_external_data(self, mock_ctx):
        """Test that external_data in metadata is merged correctly."""
        internal_data = {
            "prompt": [["user", "Test prompt"]],
            "metadata": {
                "id": "b2761a88-0662-453b-aba7-27846434dc59",
                "created_at": 1234567890,
                "external_data": {"source_dataset": "custom", "annotation_quality": "high"},
            },
        }

        temp_file = create_temp_dataset_file(internal_data, mock_ctx)
        try:
            dataset = Dataset[dataset_kinds.Prompt](dataset_key="test_external_data", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 1
            assert str(threads[0].metadata["id"]) == "b2761a88-0662-453b-aba7-27846434dc59"
            assert threads[0].metadata["source_dataset"] == "custom"
            assert threads[0].metadata["annotation_quality"] == "high"
        finally:
            os.unlink(temp_file)

    def test_load_dataset_any_format(self, mock_ctx):
        data_lines = [
            {"input": [{"role": "user", "content": "Question 1?"}], "completion": "Answer 1", "metadata": {"id": 1}},
            {"input": [{"role": "user", "content": "Question 2?"}], "completion": "Answer 2", "metadata": {"id": 2}},
        ]

        temp_file = create_temp_dataset_file(data_lines, mock_ctx)
        try:
            dataset = Dataset(dataset_key="test_external_data", local_file_path=temp_file)
            threads = asyncio.run(dataset.load(mock_ctx))

            assert len(threads) == 2
            assert all(isinstance(thread, StringThread) for thread in threads)
            assert threads[0].metadata["id"] == 1
            assert threads[1].metadata["id"] == 2
            assert "Answer 1" in threads[0].last_content()
            assert "Answer 2" in threads[1].last_content()
        finally:
            os.unlink(temp_file)



import asyncio

from adaptive_harmony import StringThread
from adaptive_harmony.core.reward_client.client import Request, RewardClient, Turn
from adaptive_harmony.graders.base_grader import BaseGrader, Grade


class RewardServerGrader(BaseGrader):
    def __init__(self, grader_key: str, grader_id: str, reward_server_ip: str):
        super().__init__(grader_key)
        self.reward_client = RewardClient(reward_server_ip)
        self.grader_id_or_key = grader_id or grader_key
        self._setup_task = None
        self._setup_lock = None

    async def _ensure_setup(self):
        if self._setup_lock is None:
            self._setup_lock = asyncio.Lock()

        if self._setup_task is None:
            async with self._setup_lock:
                if self._setup_task is None:
                    self._setup_task = asyncio.create_task(self.reward_client.setup())

        await self._setup_task

    async def grade(self, sample: StringThread) -> Grade:
        await self._ensure_setup()

        response = await self.reward_client.score(
            Request(
                turns=[Turn(content=turn.content, role=turn.role) for turn in sample.get_turns()],
                metadata=sample.metadata,
            )
        )
        return Grade(value=response.reward, grader_key=self.grader_id_or_key, reasoning=response.metadata.get("reason"))



from adaptive_harmony.core.structured_output import pydantic_parse
from adaptive_harmony.core.reasoning import remove_reasoning
from adaptive_harmony.graders.binary_judge.binary_judge import BinaryJudgeOutput


class TestRemoveReasoning:
    """Test cases for the remove_reasoning function."""

    def test_remove_simple_think_tags(self):
        """Test removal of simple think tags."""
        completion = "<think>I need to think about this carefully.</think>This is the final answer."
        expected = "This is the final answer."
        result = remove_reasoning(completion)
        assert result == expected

    def test_remove_simple_think_tags_with_multiple_tags(self):
        """Test removal of simple think tags with multiple tags (only first one removed)."""
        completion = "<think>I need to think about this carefully.</think><think>I need to think about this carefully.</think>This is the final answer."
        expected = "<think>I need to think about this carefully.</think>This is the final answer."
        result = remove_reasoning(completion)
        assert result == expected

    def test_remove_think_tags_with_newlines(self):
        """Test removal of think tags that span multiple lines."""
        completion = """<think>
This is a multi-line
thinking process
with various details.
</think>
This is the final answer."""
        expected = "This is the final answer."
        result = remove_reasoning(completion)
        assert result == expected

    def test_no_think_tags(self):
        """Test that text without think tags is unchanged."""
        completion = "This is a simple answer without any think tags."
        result = remove_reasoning(completion)
        assert result == completion

    def test_empty_think_tags(self):
        """Test removal of empty think tags."""
        completion = "<think></think>answer."
        expected = "answer."
        result = remove_reasoning(completion)
        assert result == expected

    def test_only_think_tags(self):
        """Test text that contains only think tags."""
        completion = "<think>Only thinking here</think>"
        expected = ""
        result = remove_reasoning(completion)
        assert result == expected

    def test_think_tags_with_special_characters(self):
        """Test removal of think tags containing special characters."""
        completion = "<think>Special chars: !@#$%^&*()</think>answer."
        expected = "answer."
        result = remove_reasoning(completion)
        assert result == expected

    def test_qwen_model_no_opening_tag(self):
        """Test removal of reasoning content from start to </think> (Qwen models)."""
        completion = """I need to think about this carefully:
1. First, I need to understand the criteria
2. Then I need to evaluate the completion
3. Finally, I need to provide a score

Based on my analysis, the completion meets the criteria.
</think>

This is the final answer."""
        expected = "This is the final answer."
        result = remove_reasoning(completion)
        assert result == expected

    def test_qwen_model_multiline_reasoning(self):
        """Test removal of multiline reasoning content ending with </think>."""
        completion = """Let me analyze this step by step:

The user is asking for help with their code.
I should provide a clear and helpful response.
The solution involves updating the regex pattern.
</think>

Here's the updated code you need."""
        expected = "Here's the updated code you need."
        result = remove_reasoning(completion)
        assert result == expected

    def test_structured_output_example(self):
        """Test a complete workflow example."""
        # Simulate a completion with think tags
        completion_with_thinking = """<think>
Let me think about this carefully:
1. First, I need to understand the criteria
2. Then I need to evaluate the completion
3. Finally, I need to provide a score

Based on my analysis, the completion meets the criteria.
</think>

```json
{
    "reasoning": "The completion meets the criteria and follows the guidelines",
    "score": "PASS"
}
```"""

        # Create BinaryJudgeOutput with the cleaned completion
        parsed = pydantic_parse(completion_with_thinking, BinaryJudgeOutput)

        # Verify the results
        assert parsed.score == "PASS"
        assert "meets the criteria" in parsed.reasoning



import asyncio
from adaptive_harmony import get_client, StringThread
from adaptive_harmony.graders import BaseGrader
from adaptive_harmony.graders.binary_judge import BinaryJudgeGrader, BinaryJudgeShot
from adaptive_harmony.graders.combined_grader import CombinedGrader
from adaptive_harmony.graders.faithfulness_judge import FaithfulnessGrader
from pprint import pprint


async def main():
    client = await get_client("ws://localhost:50053")

    humor_shots = [
        BinaryJudgeShot(
            thread=StringThread([("user", "tell me a joke"), ("assistant", "no")]),
            reasoning="The response is not funny, in fact it is not even a joke.",
            score="FAIL",
        )
    ]

    model = await client.model("openai://gpt-4o-mini").tp(1).spawn_inference("judge_model")
    # Humor binary judge

    humor_grader = BinaryJudgeGrader(
        grader_key="humor_grader",
        model=model,
        criteria="The response should be super funny and make people laugh.",
        shots=humor_shots,
    )

    # Conditional not related binary judge (test NA)
    conditional_grader = BinaryJudgeGrader(
        grader_key="conditional_grader",
        model=model,
        criteria="If the user is asking about car racing topics, the model should abstain from answering.",
        shots=None,
    )

    # Faithfulness scorer
    faithfulness_grader = FaithfulnessGrader(
        grader_key="faithfulness_grader",
        model=model,
        language="en",
    )

    async def example_scoringfunction(x: StringThread) -> float:
        return 1.0

    BaseGrader.from_function(grader_key="function_grader", async_fn=example_scoringfunction)

    all_graders = [humor_grader, conditional_grader, faithfulness_grader]
    # Create combined scorer
    combined_scorer = CombinedGrader(
        grader_key="combined_scorer",
        graders=all_graders,
        aggregation_method="mean",
    )

    # Generate a response to test
    prompt = "Tell me a funny joke. A funny joke to me is a joke that involves puns"
    completion = "Sure! Here's a pun-filled joke for you: I once heard a joke about a pencil… but it had no point."
    thread = StringThread().user(prompt).assistant(completion)

    print(f"Generated response:\n {thread.last_content()}")
    print()

    # Test the combined scorer
    print("=== Testing Combined Scorer ===")
    grade = await combined_scorer.grade(thread)

    print(f"Combined Score: {grade.value}")
    print()
    print("Detailed reason:")
    pprint(grade.reasoning)


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python

"""
Progress tracking and coroutine observability for async operations.

Provides real-time visualization of async task execution with a progress bar
and hierarchical coroutine call tree display.
"""

import asyncio
import os
import weakref

from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
)
from rich.table import Table
from rich.text import Text


def _is_stdlib(filename: str) -> bool:
    """Check if a filename is from Python's standard library."""
    if not filename:
        return False

    # Check for common stdlib patterns
    stdlib_indicators = [
        "asyncio",
        "threading",
    ]

    return any(indicator in filename for indicator in stdlib_indicators)


def describe_coroutine(coro):
    """
    Extracts and returns all user code locations in the await chain as a list
    (excluding stdlib). Each entry is formatted as "filename:lineno code..."
    """
    user_code_locations = []

    # Follow the cr_await chain to collect all user coroutines
    while coro is not None:
        frame = getattr(coro, "cr_frame", None)
        if frame is None:
            break

        filename = frame.f_code.co_filename
        lineno = frame.f_lineno

        # Don't descend into stdlib code
        if _is_stdlib(filename):
            user_code_locations.append("Internals:...")
            break

        # Collect this user code location (no indentation - tree will handle that)
        # Truncate long file paths to just the filename
        short_filename = filename.split("/")[-1]
        # Get function name from frame
        func_name = frame.f_code.co_name
        # Keep it very short to fit in panel
        user_code_locations.append(f"{short_filename}:{lineno} {func_name}()")

        # Check if this coroutine is awaiting another coroutine
        awaited = getattr(coro, "cr_await", None)
        if awaited is None or not hasattr(awaited, "cr_frame"):
            # This is the innermost coroutine
            break

        # Descend into the awaited coroutine
        coro = awaited

    return user_code_locations


class ProgressCounterRegistry:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if self._initialized:
            return
        self._main_counter_instance: ProgressCounter | None = None
        self._wrappers: dict[str, ProgressCounterWrapper] = {}
        self._initialized = True

    def get_main_counter(self) -> "ProgressCounter | None":
        return self._main_counter_instance

    def set_main_counter(self, counter: "ProgressCounter | None"):
        self._main_counter_instance = counter

    def get_wrapper(self, key: str) -> "ProgressCounterWrapper | None":
        return self._wrappers.get(key)

    def set_wrapper(self, key: str, wrapper: "ProgressCounterWrapper"):
        self._wrappers[key] = wrapper

    def reset(self):
        self._main_counter_instance = None
        self._wrappers.clear()


class CoroutineTree:
    """Manages building and formatting a tree structure from coroutine call chains."""

    def __init__(self):
        self.max_height = 1

    def build_from_chains(self, chains: list[list[str]]) -> dict:
        """Build a tree structure from coroutine call chains."""
        tree = {}

        for chain in chains:
            current = tree
            for desc in chain:
                if desc not in current:
                    current[desc] = {"count": 0, "children": {}}
                current[desc]["count"] += 1
                current = current[desc]["children"]

        return tree

    def format_as_text(self, tree: dict, prefix: str = "") -> Text:
        """Format tree structure as Rich Text with counts."""
        text = Text()
        items = list(tree.items())

        for i, (desc, node) in enumerate(items):
            is_last_item = i == len(items) - 1

            # Tree branch characters
            if prefix == "":
                branch = ""
            else:
                branch = "└── " if is_last_item else "├── "

            # Format line with count
            count_str = f"[{node['count']}] "
            text.append(prefix + branch, style="dim")
            text.append(count_str, style="bold magenta")
            text.append(desc + "\n")

            # Recurse into children
            if node["children"]:
                # Add indentation for children
                new_prefix = prefix + ("    " if is_last_item else "│   ")
                text.append(self.format_as_text(node["children"], new_prefix))

        return text

    def create_padded_display(self, chains: list[list[str]], active_count: int) -> Text:
        """Create a formatted and padded display text for the tree."""
        if not chains:
            return Text("No chains available", style="dim")

        # Build and format tree
        tree = self.build_from_chains(chains)
        header = Text(f"{active_count} tasks active\n\n", style="bold cyan")
        tree_content = self.format_as_text(tree)

        # Count lines to track height
        current_height = len(header.plain.split("\n")) + len(tree_content.plain.split("\n"))
        if current_height > self.max_height:
            self.max_height = current_height

        # Combine and add padding
        display = Text()
        display.append(header)
        display.append(tree_content)

        # Add padding to maintain consistent height
        padding_lines = max(0, self.max_height - current_height)
        for _ in range(padding_lines):
            display.append("\n")

        return display

    def create_empty_display(self) -> Text:
        """Create a display for when there are no active tasks."""
        display = Text("No active tasks but updated", style="dim")

        # Pad to maintain consistent height
        current_height = 1
        padding_lines = max(0, self.max_height - current_height)
        for _ in range(padding_lines):
            display.append("\n")

        return display


class ProgressCounter:
    def __init__(self, main_job_name: str, total_tasks: int):
        registry = ProgressCounterRegistry()
        assert registry.get_main_counter() is None, "Only one main counter instance is allowed"
        self.total_tasks = total_tasks

        # construct the left panel that will give the state of the batch at a glance
        self.overall_progress = Progress(
            "{task.description}",
            BarColumn(),
            TextColumn("[magenta]{task.completed:n} steps"),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TimeRemainingColumn(),
        )
        self.overall_task = self.overall_progress.add_task(main_job_name, total=self.total_tasks)
        self.overall_panel = Panel(
            self.overall_progress, title="Overall Progress", border_style="green", padding=(1, 1, 0, 1)
        )

        # construct the right panel for the coroutine tree
        self.coroutine_tree = CoroutineTree()
        self.tree_text = Text("Waiting for tasks...", style="dim")
        self.tree_panel = Panel(self.tree_text, title="[b]Coroutine Tree", border_style="blue", padding=(1, 2))

        self._monitored_tasks: weakref.WeakSet[asyncio.Task] = weakref.WeakSet()
        self._monitor_task: asyncio.Task | None = None
        self.interval = 0.5

        # Construct the general table with two columns
        self.progress_table = Table.grid()
        self.progress_table.add_row(self.overall_panel, self.tree_panel)

        # Disable the display if DISABLE_RICH_PROGRESS is set
        # TODO: make this less dirty
        self.enable_display = not os.getenv("DISABLE_RICH_PROGRESS", "0") == "1"

    def register_task(self, task: asyncio.Task):
        self._monitored_tasks.add(task)

    def increment_total_counter(self):
        self.overall_progress.advance(TaskID(0))

    def is_done(self):
        return self.overall_progress.finished

    async def _monitor_loop(self):
        try:
            while True:
                await asyncio.sleep(self.interval)

                # Filter out done tasks (just in case gc hasn't run yet)
                active = [t for t in self._monitored_tasks if not t.done()]

                # Collect all call chains
                chains = []
                for task in active:
                    coro = task.get_coro()
                    descriptions = describe_coroutine(coro)
                    if descriptions:
                        chains.append(descriptions)

                # Create tree display
                if not active or not chains:
                    self.tree_text = self.coroutine_tree.create_empty_display()
                else:
                    self.tree_text = self.coroutine_tree.create_padded_display(chains, len(active))

                # Update tree panel
                self.tree_panel = Panel(self.tree_text, title="[b]Coroutine Tree", border_style="blue", padding=(1, 2))

                # Update overall panel with matching height
                overall_with_padding = Group(
                    self.overall_progress,
                    Text("\n" * max(0, self.coroutine_tree.max_height - 2)),  # Account for progress bar base height
                )
                self.overall_panel = Panel(
                    overall_with_padding, title="Overall Progress", border_style="green", padding=(1, 1, 0, 1)
                )

                # Recreate the progress table with updated panels
                self.progress_table = Table.grid()
                self.progress_table.add_row(self.overall_panel, self.tree_panel)
                if self.enable_display:
                    self.live.update(self.progress_table)

        except asyncio.CancelledError:
            pass

    async def __aenter__(self):
        if self.enable_display:
            self.live = Live(self.progress_table, refresh_per_second=4).__enter__()
        self._monitor_task = asyncio.create_task(self._monitor_loop())
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        # Cancel the monitor task
        if self._monitor_task and not self._monitor_task.done():
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass

        if self.enable_display:
            self.live.__exit__(exc_type, exc_value, exc_traceback)
        registry = ProgressCounterRegistry()
        assert registry.get_main_counter() is not None, "Weird state"
        # we clear both the main counter and any wrapper that would point to the old counter
        registry.reset()


class ProgressCounterWrapper(ProgressCounter):
    """Used to wrap a preexisting counter, needed to have simple logic in case an async_map is used inside an async_map_batch"""

    def __init__(self, inner: ProgressCounter, main_function_name: str):
        self.inner = inner
        self.main_function_name = main_function_name

    def register_task(self, task: asyncio.Task):
        return self.inner.register_task(task)

    def increment_total_counter(self):
        # For now we do nothing
        ...

    async def __aenter__(self):
        # do nothing, the inner progress counter has already been initialized
        return self

    async def __aexit__(self, _exc_type, _exc_value, _exc_traceback):
        # do nothing, the inner progress counter will be exited later
        ...


def get_progress_counter_or_wrapper(main_job_name: str, total_samples: int):
    registry = ProgressCounterRegistry()
    main_counter = registry.get_main_counter()

    if main_counter is None:
        main_counter = ProgressCounter(main_job_name, total_samples)
        registry.set_main_counter(main_counter)
        return main_counter
    else:
        wrapper = registry.get_wrapper(main_job_name)
        if wrapper is None:
            wrapper = ProgressCounterWrapper(main_counter, main_job_name)
            registry.set_wrapper(main_job_name, wrapper)
        return wrapper



from harmony_client.artifacts import CustomArtifact as CustomArtifact
from harmony_client.artifacts import DatasetArtifact as DatasetArtifact

from adaptive_harmony.evaluation import EvaluationArtifact as EvaluationArtifact



# Re-export from harmony_client.runtime.runner
from harmony_client.runtime.context import RecipeConfig as RecipeConfig  # noqa: F401
from harmony_client.runtime.context import RecipeContext as RecipeContext
from harmony_client.runtime.runner import (  # noqa: F401
    RunnerArgs as RunnerArgs,
)
from harmony_client.runtime.runner import (
    _download_and_extract_recipe as _download_and_extract_recipe,
)
from harmony_client.runtime.runner import (
    _get_params as _get_params,
)
from harmony_client.runtime.runner import (
    _install_recipe_dependencies as _install_recipe_dependencies,
)
from harmony_client.runtime.runner import (
    _load_and_run_recipe as _load_and_run_recipe,
)
from harmony_client.runtime.runner import (
    _parse_script_metadata as _parse_script_metadata,
)
from harmony_client.runtime.runner import (
    main as main,
)

if __name__ == "__main__":
    main()



import logging
import uuid
from typing import List, Self

from harmony_client import (
    EvalSample,
    EvaluationArtifactBase,
)
from harmony_client.runtime.context import RecipeContext

logger = logging.getLogger(__name__)


class EvaluationArtifact:
    def __init__(self, name: str, ctx: RecipeContext) -> None:
        artifact_id = str(uuid.uuid4())
        url = ctx.file_storage.mk_url(f"artifacts/eval_samples_{artifact_id}.jsonl")
        self._base = EvaluationArtifactBase(name, url, artifact_id)
        self.ctx = ctx
        self.ctx.job.register_artifact(self._base.artifact)

    @property
    def id(self) -> str:
        return self._base.id

    @property
    def name(self) -> str:
        return self._base.name

    @property
    def kind(self) -> str:
        return self._base.kind

    @property
    def uri(self) -> str:
        assert self._base.uri is not None
        return self._base.uri

    def add_samples(self, samples: List[EvalSample]) -> Self:
        """Add evaluation samples to this artifact.

        Args:
            samples: List of evaluation samples to add

        Returns:
            Self for method chaining

        Raises:
            ValueError: If samples list is empty
            Exception: If serialization or storage fails
        """
        if not samples:
            raise ValueError("Cannot add empty samples list")

        try:
            samples_json = self._base.samples_to_adaptive_json(samples)
            for json_str in samples_json:
                self.ctx.file_storage.append((json_str + "\n").encode("utf-8"), self.uri)
            logger.debug(f"Added {len(samples)} samples to artifact {self.id}")
        except Exception as e:
            logger.error(f"Failed to add samples to artifact {self.id}: {e}")
            raise

        return self

    def __repr__(self):
        return f"EvaluationArtifact(id={self.id}, name={self.name}, kind={self.kind}, uri={self.uri})"



SYSTEM = """Your task is to judge the faithfulness of a series of statements to the provided context.
For each statement_idx you must return your reasoning to support a score you attribute to the corresponding statement.
The score should be 1 in case the statement is fully supported and can be directly inferred based on the context, or 0 in case it cannot.
If there is no relevant context to be faithful to, the score should be 1.
You must score every single sentence without skipping any.

If it exists, you will be given the whole CONVERSATION so far leading up to the LAST USER TURN.
You must evaluate the statements with a focus on what is being asked in the LAST USER TURN, and never on an intermediary questions that might have been asked in course of the conversation.

You always output a JSON object with the following schema, and nothing else before or after:
{json_schema}"""


USER = """CONVERSATION
{context}

LAST USER TURN
{user_question}

STATEMENTS
{sentences}
"""



from abc import abstractmethod
from typing import Any

import numpy as np
from harmony_client import (
    InferenceModel,
    StringThread,
    TrainingModel,
)
from loguru import logger

from adaptive_harmony.core.utils import async_map, async_map_fallible
from adaptive_harmony.environment import EnvironmentFactory
from adaptive_harmony.graders import BaseGrader
from adaptive_harmony.logging_table import Table


class RecipeCallback:
    def __init__(self, frequency: float, log_key_prefix: str | None = None):
        self.frequency = frequency
        self.last_call = -1.0
        self.log_key_prefix = log_key_prefix

    async def maybe_call(self, current_percentage: float) -> dict[str, Any]:
        if current_percentage - self.last_call >= self.frequency:
            self.last_call = current_percentage
            callback_dict = await self.callback(current_percentage)
            prefixed_dict = {
                (f"{self.log_key_prefix}/{key}" if self.log_key_prefix else key): value
                for key, value in callback_dict.items()
            }
            return prefixed_dict
        return {}

    @abstractmethod
    async def callback(self, current_percentage: float) -> dict[str, Any]: ...


class GenerateSamplesCallback(RecipeCallback):
    def __init__(
        self,
        thread_set: list[StringThread],
        model: InferenceModel,
        frequency: float,
        log_key: str = "samples",
    ):
        super().__init__(frequency, log_key_prefix="generation")
        self.thread_set = thread_set
        self.model = model
        self.log_key = log_key

    async def callback(self, current_percentage: float) -> dict[str, Any]:
        logger.info("Entering generation callback...")
        generation_tokens = await async_map_fallible(self.model.generate_tokens, self.thread_set)
        generation_results = await async_map_fallible(self.model.detokenize_thread, generation_tokens)
        gen_lengths = [sample.len_last_turn() for sample in generation_tokens]

        generation_logs = {
            self.log_key: Table()
            .add_column(
                "system",
                [
                    sample.get_turns()[0].content if sample.get_turns()[0].role == "system" else ""
                    for sample in generation_results
                ],
            )
            .add_column(
                "prompt",
                [
                    repr(
                        StringThread(
                            sample.get_turns()[1:-1]
                            if (sample.get_turns() and sample.get_turns()[0].role == "system")
                            else sample.get_turns()[:-1]
                        )
                    )
                    for sample in generation_results
                ],
            )
            .add_column("response", [response.last_content() for response in generation_results]),
            "generation_length_mean": np.mean(gen_lengths).item(),
            "generation_length_std": np.std(gen_lengths).item(),
            "num_samples": len(generation_results),
        }
        return generation_logs


class ValidationLossCallback(RecipeCallback):
    def __init__(
        self,
        validation_set: list[StringThread],
        model: InferenceModel,
        frequency: float = 0.1,
        log_key: str = "loss",
    ):
        super().__init__(frequency, log_key_prefix="validation")
        self.validation_set = validation_set
        self.model = model
        self.log_key = log_key

    async def callback(self, current_percentage: float) -> dict[str, float]:
        logger.info("Entering validation loss callback...")
        losses = []
        tokens = await async_map_fallible(self.model.tokenize_thread, self.validation_set)
        logprobs = await async_map(self.model.logprobs_per_token, tokens)
        losses = [-(sum(lp) / len(lp)) for lp in logprobs]

        return {self.log_key: sum(losses) / len(losses)}


class CheckpointCallback(RecipeCallback):
    def __init__(
        self,
        model: TrainingModel,
        checkpoint_name: str,
        frequency: float = 0.2,
    ):
        super().__init__(frequency, log_key_prefix="checkpointing")
        self.last_call = 0.0  # avoid saving the model at the first period
        self.model = model
        self.model_log_name = checkpoint_name

    async def callback(self, current_percentage: float):
        logger.info(f"Saving checkpoint at {current_percentage * 100} % of training ...")
        await self.model.save(f"{self.model_log_name}-{round(current_percentage, 3)}")
        return {}


class GraderEvalCallback(RecipeCallback):
    def __init__(
        self,
        validation_set: list[StringThread],
        model: InferenceModel,
        grader: BaseGrader,
        frequency: float,
        log_key: str = "validation",
        clear_grader_logs: bool = True,
        temperature: float = 0.0,
    ):
        super().__init__(frequency, log_key_prefix=log_key)
        self.validation_set = validation_set
        self.model = model
        self.grader = grader
        self.clear_grader_logs = clear_grader_logs
        self.temperature = temperature

    async def callback(self, current_percentage: float) -> dict[str, float | Table]:
        logger.info("Entering grader evaluation callback...")
        temp_model = self.model.temperature(self.temperature)

        tokenized_results = await async_map_fallible(temp_model.generate_tokens, self.validation_set)
        string_results = await async_map(temp_model.detokenize_thread, tokenized_results)
        grades = await async_map_fallible(self.grader.grade, string_results)
        gen_lengths = [sample.len_last_turn() for sample in tokenized_results]

        grader_logs = self.grader.get_logs(clear=self.clear_grader_logs)
        return {
            **{f"rewards/{key}": value for key, value in grader_logs.items()},
            "generation_length_mean": float(np.mean(gen_lengths).item()),
            "generation_length_std": float(np.std(gen_lengths).item()),
            "num_samples": float(len(grades)),
        }


class EnvironmentValidationCallback(RecipeCallback):
    def __init__(
        self,
        validation_set: list[StringThread],
        model: InferenceModel,
        env_factory: EnvironmentFactory,
        frequency: float,
        log_key: str = "validation",
        clear_env_logs: bool = True,
        temperature: float = 0.0,
        num_samples_log: int = 0,
    ):
        super().__init__(frequency, log_key_prefix=log_key)
        self.validation_set = validation_set
        self.model = model
        self.env_factory = env_factory
        self.clear_env_logs = clear_env_logs
        self.temperature = temperature
        self.num_samples_log = num_samples_log

    async def generate_trajectory(self, initial_thread: StringThread) -> tuple[StringThread, float, int]:
        env = self.env_factory.create_environment(initial_thread.metadata)
        temp_model = self.model.temperature(self.temperature)
        trajectory, trajectory_score = await env.generate_trajectory_and_grade(temp_model, initial_thread)
        num_turns = len([turn for turn in trajectory.get_turns() if turn.role == "assistant"])
        return trajectory, trajectory_score.cumulative_score, num_turns

    async def callback(self, current_percentage: float) -> dict[str, float | Table]:
        logger.info("Entering environment validation callback...")

        results = await async_map_fallible(self.generate_trajectory, self.validation_set)

        trajectories = [traj for traj, _, _ in results]
        scores = [score for _, score, _ in results]
        num_turns_list = [num_turns for _, _, num_turns in results]

        validation_logs = {
            "score_mean": np.mean(scores).item(),
            "score_std": np.std(scores).item(),
            "num_turns_mean": np.mean(num_turns_list).item(),
            "num_turns_std": np.std(num_turns_list).item(),
            "num_samples": len(results),
        }

        env_logs = self.env_factory.get_logs(clear=self.clear_env_logs)
        validation_logs.update({f"env/{key}": value for key, value in env_logs.items()})

        if self.num_samples_log > 0:
            samples = [repr(traj) for traj in trajectories[: self.num_samples_log]]
            samples_scores = scores[: self.num_samples_log]
            table = Table().add_column("trajectory", samples).add_column("score", samples_scores)
            validation_logs["samples"] = table

        logger.info(f"Validation Mean score: {validation_logs['score_mean']:.4f}")
        return validation_logs



from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from numbers import Number
from typing import Any, Callable

import numpy as np

from adaptive_harmony import StringThread
from adaptive_harmony.logging_table import Table


@dataclass
class TurnScore:
    score: float
    num_assistant_turns: int


@dataclass
class TrajectoryScore:
    scores: list[TurnScore]

    def to_turn_scores(self) -> list[float]:
        return [turn_score.score for turn_score in self.scores for _ in range(turn_score.num_assistant_turns)]

    @property
    def cumulative_score(self) -> float:
        return sum([turn_score.score for turn_score in self.scores])


class Environment(ABC):
    """
    Environment to inherit from when building an environment.
    """

    def __init__(self, logging_function: Callable):
        self.logging_function = logging_function

    @abstractmethod
    async def react_to(self, thread: StringThread) -> list[tuple[str, str]] | TrajectoryScore:
        """Returns either [("tool", tool response), ...] or [("user", user question)] or TrajectoryScore when DONE."""
        pass

    async def bootstrap_prompt(self, thread: StringThread) -> StringThread:
        return thread

    async def generate_trajectory_and_grade(
        self, model, initial_thread: StringThread
    ) -> tuple[StringThread, TrajectoryScore]:
        """Generate a full trajectory by interacting with the model until termination."""
        thread = await self.bootstrap_prompt(initial_thread)

        while True:
            # Generate model response
            thread = await model.generate(thread)

            # Get environment reaction
            env_response = await self.react_to(thread)

            # If we got a score, we're done
            if isinstance(env_response, TrajectoryScore):
                return thread, env_response

            # Otherwise, add the environment responses to the thread
            for role, content in env_response:
                if role == "tool":
                    thread = thread.tool(content)
                elif role == "user":
                    thread = thread.user(content)
                else:
                    raise ValueError(f"Unknown role: {role}")


class EnvironmentFactory(ABC):
    """
    Abstract class to build environments. It is necessary because each trajectory must have its own unique Environment
    """

    def __init__(self, logging_name: str | None = None):
        self._logs: dict[str, list] = defaultdict(list)
        self.logging_name = logging_name

    def add_log(self, key, new_value) -> None:
        """Add a log entry to the scorer's log collection."""
        self._logs[key].append(new_value)

    def get_logs(self, clear: bool = False) -> dict[str, float | Table]:
        """
        Get aggregated logs from all score calls.
        Base implementation computes statistics for "score" keys in individual logs.
        If there are none, returns empty dict.
        """
        if not self._logs:
            return {}

        logs = {}
        for k, v in self._logs.items():
            if isinstance(v[0], Number):
                logs[k] = np.asarray(v).mean()
            elif isinstance(v[0], Table):
                headers = v[0].headers
                assert all(table.headers == headers for table in v)
                overall_table = Table(headers)
                for table in v:
                    overall_table.add_rows(table.rows)
                logs[k] = overall_table
            else:
                raise ValueError(f"Unknown type: {type(v[0])}")

        if clear:
            self.clear_logs()
        return logs

    def clear_logs(self) -> None:
        """
        Clear all accumulated logs.
        """
        self._logs.clear()

    @abstractmethod
    def create_environment(self, metadata: Any) -> Environment: ...



from dataclasses import dataclass, field


@dataclass
class ResponseAccumulator:
    total_num_chunks: int = 0
    chunks_received: list[str] = field(default_factory=list)

    def add_chunk(self, chunk: str):
        self.chunks_received.append(chunk)

    def is_complete(self) -> bool:
        return len(self.chunks_received) == self.total_num_chunks

    def get_full_data(self) -> str | None:
        if self.is_complete():
            return "".join(self.chunks_received)
        return None



import asyncio

import numpy as np

from adaptive_harmony import InferenceModel, StringThread
from adaptive_harmony.core.structured_output import JsonParseError
from adaptive_harmony.core.utils import stringify_thread
from adaptive_harmony.graders import BaseGrader, Grade
from adaptive_harmony.graders.range_judge.prompts import (
    RangeJudgeShot,
    RangeScorerTemplates,
    ReasonedScore,
    SubrangeExpectations,
    create_shots,
    get_prompt_building_blocks,
)
from adaptive_harmony.graders.utils import (
    FailedJudgeLog,
    SuccessJudgeLog,
    sample_score_distribution,
)
from adaptive_harmony.logging_table import Table


class RangeJudgeGrader(BaseGrader[SuccessJudgeLog | FailedJudgeLog]):
    """
    Scores a thread in a range of integer scores, based on a list of evaluation steps.
    If evaluation steps are not provided, they are generated from the criteria.
    The final score is computed as a weighted average of all possible scores,
    where the weights are the logprobs of each score.
    You can pass subrange_expectations to the scorer, to help the judge
    understand the correspondence between score subranges and expected quality levels.
    """

    def __init__(
        self,
        grader_key: str,
        model: InferenceModel,
        criteria: str,
        score_range: tuple[int, int] = (1, 5),
        evaluation_steps: list[str] | None = None,
        subrange_expectations: list[SubrangeExpectations] | None = None,
        shots: list[RangeJudgeShot] | None = None,
        normalize_score: bool = True,
    ):
        model_path: str = model.get_builder_args().get("path")  # type: ignore[assignment]
        assert model_path.startswith("model_registry://"), "External models cannot be used in RangeJudgeScorer"

        super().__init__(grader_key)
        self.model = model
        self.criteria = criteria
        self.score_range = score_range
        self.min_score, self.max_score = score_range
        self.subrange_expectations = subrange_expectations
        self._shots = shots
        self.normalize_score = normalize_score

        if evaluation_steps is None:
            self._str_evaluation_steps = None
            self._list_eval_steps = None
        else:
            self._str_evaluation_steps = "\n".join([f"{i + 1}: {step}" for i, step in enumerate(evaluation_steps)])
            self._list_eval_steps = evaluation_steps

    @property
    def evaluation_steps(self) -> list[str] | None:
        return self._list_eval_steps

    @property
    def str_evaluation_steps(self) -> str | None:
        return self._str_evaluation_steps

    @evaluation_steps.setter
    def evaluation_steps(self, steps: list[str]):
        self._list_eval_steps = steps
        self._str_evaluation_steps = "\n".join([f"{i + 1}: {step}" for i, step in enumerate(steps)])

    async def generate_evaluation_steps(self) -> list[str]:
        thread = await self.model.temperature(0.0).generate(RangeScorerTemplates.get_evaluation_steps(self.criteria))
        self._str_evaluation_steps = thread.last_content()
        self._list_eval_steps = self._str_evaluation_steps.split("\n")
        assert self.evaluation_steps
        return self.evaluation_steps

    async def grade(self, sample: StringThread) -> Grade:
        if self.evaluation_steps is None:
            if self._shots is not None:
                raise ValueError(
                    "You cannot pass shots without specifying evaluations steps, since your shots' reasoning must match the steps"
                )
            await self.generate_evaluation_steps()

        # Format shots into both user turn formats (for first reasoned scoring step, and for last logprobs scoring step)
        assert self.str_evaluation_steps

        shots = create_shots(self.criteria, self.str_evaluation_steps, self._shots) if self._shots is not None else {}

        # Separate relevant parts of the prompt turns
        prompt_components = get_prompt_building_blocks(sample)
        # Get reasoned scoring thread
        eval_thread = RangeScorerTemplates.get_json_reasoned_score(
            context=prompt_components.context,
            last_user_input=prompt_components.last_user_turn,
            assistant_answer=prompt_components.last_assistant_turn,
            criteria=self.criteria,
            evaluation_steps=self.str_evaluation_steps,
            score_range=self.score_range,
            json_schema=self.model.render_schema(ReasonedScore),
            subrange_expectations=self.subrange_expectations,
            shots=shots.get("reasoning"),
        )
        eval_str_prompt = stringify_thread(eval_thread, sep=f"\n\n{'-' * 10}\n\n")
        try:
            _, reasoned_score = await self.model.temperature(0.0).generate_and_validate(eval_thread, ReasonedScore)
        except JsonParseError as e:
            self.add_log({"prompt": eval_str_prompt, "error": f"{str(e)}\n\nCOMPLETION:\n{e.completion}"})
            raise
        except Exception as e:
            self.add_log({"prompt": eval_str_prompt, "error": str(e)})
            raise

        # Get a prompt that includes the reasoning for the sample, all the way to form-filling the score
        up_to_score_thread = RangeScorerTemplates.get_up_to_score(
            context=prompt_components.context,
            last_user_input=prompt_components.last_user_turn,
            assistant_answer=prompt_components.last_assistant_turn,
            criteria=self.criteria,
            evaluation_steps=self.str_evaluation_steps,
            score_range=self.score_range,
            reasoning=reasoned_score.reasoning,
            subrange_expectations=self.subrange_expectations,
            shots=shots.get("scoring"),
        )

        # Get logprobs for each possible final score
        possible_score_ints = [s for s in range(self.min_score, self.max_score + 1)]
        logprobs = await asyncio.gather(
            *[self.model.temperature(0.0).logprobs(up_to_score_thread.assistant(f"{s}")) for s in possible_score_ints]
        )

        # Convert to probabilities and compute weighted average
        probs = np.exp(logprobs - np.logaddexp.reduce(logprobs))
        weighted_score = np.average(possible_score_ints, weights=probs)

        final_score: float = weighted_score
        if self.normalize_score:  # normalize to 0-1 range
            final_score = (weighted_score - self.min_score) / (self.max_score - self.min_score)

        str_prompt = stringify_thread(eval_thread, sep=f"\n\n{'-' * 10}\n\n")
        self.add_log({"score": final_score, "prompt": str_prompt, "reasoning": reasoned_score.reasoning})

        metadata = dict(
            criteria=self.criteria,
            raw_avg_score=float(weighted_score),
            scale_range=(self.min_score, self.max_score),
            score_probabilities={str(score): float(prob) for score, prob in zip(possible_score_ints, probs)},
            evaluation_steps=self.evaluation_steps,
            reasoning=reasoned_score.reasoning,
        )

        return Grade(value=float(final_score), grader_key=self.grader_key, reasoning=reasoned_score.reasoning)

    def add_log(self, log_data: SuccessJudgeLog | FailedJudgeLog) -> None:
        self._logs.append(log_data)

    def get_logs(self, clear: bool = False, log_all_samples: bool = False) -> dict[str, float | Table]:
        # Only clear logs at the end if clear is True
        logs = super().get_logs(clear=False)

        successfully_scored_samples = [log for log in self._logs if "score" in log]

        # stratified sample range of scores to see high and low
        if not log_all_samples:
            subset_successfully_scored_samples = sample_score_distribution(successfully_scored_samples, 15)
        else:
            # if we have fewer than 15 samples or we want to log all samples, take them all
            subset_successfully_scored_samples = successfully_scored_samples

        failed_scored_samples = [log for log in self._logs if "error" in log]

        sample_logs = self.get_sample_tables(subset_successfully_scored_samples, failed_scored_samples)

        logs.update(sample_logs)

        if clear:
            self.clear_logs()

        return logs



import asyncio
import functools
import hashlib
import itertools
import json
import random
from pathlib import Path
from typing import Any, Callable, Coroutine, Dict, Iterator, List, NamedTuple, Sequence, TypedDict, TypeVar

import numpy as np
from loguru import logger

from adaptive_harmony import InferenceModel, StringThread, TrainingModel
from adaptive_harmony.core.rich_counter import ProgressCounter, get_progress_counter_or_wrapper
from adaptive_harmony.metric_logger import Logger, StdoutLogger

S = TypeVar("S")
T = TypeVar("T")


async def wrap_coroutine_with_progress[T](coroutine: Coroutine[Any, Any, T], progress_counter: ProgressCounter) -> T:
    try:
        return await coroutine
    finally:
        progress_counter.increment_total_counter()


async def async_map_batch[S, T](
    f: Callable[[S], Coroutine[Any, Any, T]],
    data: Iterator[S],
    batch_size: int,
    max_failure_fraction: float = 0.5,
) -> List[T]:
    """
    Process items from an iterator in batches using concurrent coroutines.

    This function processes items from an iterator in batches, executing the
    provided coroutine function concurrently for each item. It excludes failing
    samples until it can create a new batch of results of size # batch size.
    If more than max_failure_fraction % of # batch size tasks fail in the process
    of creating a new batch, the function will raise the last exception encountered.
    Results are not ordered.

    Args:
        f: Coroutine function to apply to each item
        data: Iterator of items to process
        batch_size: Number of items to process in each batch

    Returns:
        List of results from successful task executions

    Note:
        - Failed tasks are not retried
        - If more than max_failure_fraction of # batch size tasks fail, the function fails
        - Tasks are automatically cancelled if the function exits early
    """
    batch_items_from_iterator = list(itertools.islice(data, batch_size))
    num_items = len(batch_items_from_iterator)

    async with get_progress_counter_or_wrapper(f"async_map_batch({f.__name__})", batch_size) as counter:
        final_results: list[Any] = [None] * num_items
        active_tasks_this_batch: Dict[asyncio.Task, int] = {}

        num_retries = 0

        for i, item_value in enumerate(batch_items_from_iterator):
            task: asyncio.Task[T] = asyncio.create_task(wrap_coroutine_with_progress(f(item_value), counter))
            counter.register_task(task)
            active_tasks_this_batch[task] = i

        try:
            while active_tasks_this_batch:
                done_tasks, _ = await asyncio.wait(active_tasks_this_batch.keys(), return_when=asyncio.FIRST_COMPLETED)

                for task_item in done_tasks:
                    original_batch_slot_idx = active_tasks_this_batch.pop(task_item)

                    try:
                        result: T = await task_item
                        final_results[original_batch_slot_idx] = result
                    except Exception as ex:
                        try:
                            if num_retries > batch_size * max_failure_fraction:
                                # if more than 50% of a batch fail we'll just go on.
                                raise ex

                            logger.debug(ex)
                            retry_item_value: S = next(data)
                            new_retry_task: asyncio.Task[T] = asyncio.create_task(
                                wrap_coroutine_with_progress(f(retry_item_value), counter)
                            )
                            active_tasks_this_batch[new_retry_task] = original_batch_slot_idx
                            num_retries += 1
                        except StopIteration:
                            ...
        finally:
            tasks_to_cancel = list(active_tasks_this_batch.keys())
            for task_to_cancel in tasks_to_cancel:
                task_to_cancel.cancel()

            if tasks_to_cancel:
                await asyncio.gather(*tasks_to_cancel, return_exceptions=True)

        if num_retries > 0:
            print(f"WARNING: had to retry {num_retries} times to get a batch of {batch_size}")
        ret = [res for res in final_results if res is not None]

        print(f"Final number tasks with non-None results: {len(ret)}")

        return ret


def hash_hyperparams(include: set[str]):
    """
    A decorator that computes a hash of specified hyperparameters and stores it on `self._hyperparams_hash`.

    Must be used on an `__init__` method. Only parameters listed in `include` will be hashed.
    Non-serializable values are converted to their string representation.

    Args:
        include: Set of parameter names to include in the hash.

    Example:
        @hash_hyperparams(include={"lr", "batch_size", "epochs"})
        def __init__(self, lr, batch_size, epochs, logger, ...):
            ...
    """
    import inspect

    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            sig = inspect.signature(func)
            bound_args = sig.bind_partial(self, *args, **kwargs)
            bound_args.apply_defaults()
            all_args = bound_args.arguments

            hyperparams = {}
            for key in include:
                if key in all_args:
                    value = all_args[key]
                    try:
                        json.dumps(value)
                        hyperparams[key] = value
                    except (TypeError, OverflowError):
                        hyperparams[key] = repr(value)

            serialized = json.dumps(hyperparams, sort_keys=True)
            self._hyperparams_hash = hashlib.sha256(serialized.encode()).hexdigest()[:16]

            return func(self, *args, **kwargs)

        return wrapper

    return decorator


def log_args(func):
    """
    A Python decorator that logs the arguments of the decorated function
    to experiment tracking tools (wandb, mlflow) or stdout.

    Attempts to log to wandb if available and initialized, then to mlflow
    if available and has an active run. If neither is available, logs to stdout.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        import inspect

        # Helper to check serializability and prepare value
        def prepare_value(value):
            # we need to log the model builder args here because they are not serializable by default
            if isinstance(value, list) and len(value) > 100:
                # exclude long lists since we want to skip datasets
                return None
            if isinstance(value, InferenceModel) or isinstance(value, TrainingModel):
                return value.get_builder_args()  # type: ignore PyRight being dumb
            else:
                # Check if the value itself is a complex object that might not be fully serializable
                try:
                    json.dumps({"test_key": value})
                    return value
                except (TypeError, OverflowError):
                    return None

        # Get function arguments once
        sig = inspect.signature(func)
        bound_args = sig.bind_partial(*args, **kwargs)
        bound_args.apply_defaults()
        all_args = bound_args.arguments

        # find the loggers that are given to recipe, if None are found, we will log to stdout
        loggers = [v for v in all_args.values() if isinstance(v, Logger)]
        if not loggers:
            loggers.append(StdoutLogger())

        # get loggable args only
        loggable_args = {k: new_v for k, v in all_args.items() if (new_v := prepare_value(v)) is not None}

        for logger_instance in loggers:
            logger_instance.log_config(loggable_args)

        return func(*args, **kwargs)

    return wrapper


async def async_map[S, T](f: Callable[[S], Coroutine[Any, Any, T]], data: Sequence[S]) -> list[T]:
    """
    Process all items in an iterable concurrently using the provided coroutine function.

    Args:
        f: Coroutine function to apply to each item
        data: Iterable of items to process

    Returns:
        List of results from all task executions
    """

    # Check if a Progress bar is already active
    async with get_progress_counter_or_wrapper(f"async_map({f.__name__})", len(list(data))) as counter:
        all_tasks = [asyncio.create_task(wrap_coroutine_with_progress(f(item), counter)) for item in data]
        for t in all_tasks:
            counter.register_task(t)
        results = await asyncio.gather(*all_tasks)
    return results


async def async_map_fallible[S, T](f: Callable[[S], Coroutine[Any, Any, T]], data: Sequence[S]) -> list[T]:
    """
    Process all items in an iterable concurrently using the provided coroutine function.

    Args:
        f: Coroutine function to apply to each item
        data: Iterable of items to process

    Returns:
        List of results from all task executions
    """

    async def wrap_coroutine_with_error_handling(coro: Coroutine[Any, Any, T]) -> tuple[T, bool]:
        try:
            result = await coro
            return result, True
        except Exception:
            return None, False  # type: ignore

    async with get_progress_counter_or_wrapper(f"async_map_fallible({f.__name__})", len(list(data))) as counter:
        all_tasks = [
            asyncio.create_task(wrap_coroutine_with_error_handling(wrap_coroutine_with_progress(f(item), counter)))
            for item in data
        ]
        for t in all_tasks:
            counter.register_task(t)
        results = await asyncio.gather(*all_tasks)

    return [result for result, success in results if success]


def get_minibatches[T](dataset: list[T], mini_batch_size: int, number_of_epochs: int) -> list[list[T]]:
    all_batches: list[list[T]] = []

    for _ in range(number_of_epochs):
        shuffled_dataset = random.sample(dataset, k=len(dataset))

        epoch_batches: list[list[T]] = []
        for i in range(0, len(shuffled_dataset), mini_batch_size):
            batch = shuffled_dataset[i : i + mini_batch_size]
            epoch_batches.append(batch)
        all_batches.extend(epoch_batches)

    return all_batches


def sample_data[T](data: list[T], epochs: float) -> list[T]:
    num_samples = len(data) * epochs
    return [data[x] for x in np.random.permutation(len(data))[: int(num_samples)]]


def weighted_mean(values: list[list[float]], weights: list[list[float]]) -> float:
    return np.average(np.concatenate(values), weights=np.concatenate(weights)).item()


def stringify_thread(thread: StringThread, sep: str = "\n\n") -> str:
    """Convert StringThread to readable text format."""
    turns = thread.get_turns()
    return sep.join([f"[{turn.role}]\n{turn.content}" for turn in turns])


class SingleTurnShot(TypedDict):
    user: dict[str, str]
    assistant: dict[str, str]


class TurnTemplates(NamedTuple):
    system: str | None
    user: str | None
    assistant: str | None
    shots: list[SingleTurnShot] | None


def turn_templates_from_dir(root_dir: str) -> TurnTemplates:
    """
    Returns system, user and assistant turn string templates from a directory, as well as a list of shot dicts.
    Expects files to be named system.md, user.md, assistant.md and shots.jsonl.
    Returns None for any turn template file that does not exist.
    """
    root_path = Path(root_dir)
    expected_files = ["system.md", "user.md", "assistant.md", "shots.jsonl"]
    missing_templates = []
    turn_templates: list[str | list[SingleTurnShot] | None] = []

    for file in expected_files:
        path = root_path / file
        if not path.exists():
            missing_templates.append(file)
            turn_templates.append(None)
        else:
            if file == "shots.jsonl":
                shots = []
                for line in path.read_text().splitlines():
                    data = json.loads(line)
                    shot = SingleTurnShot(user=data["user"], assistant=data["assistant"])
                    shots.append(shot)
                turn_templates.append(shots)
            else:
                turn_templates.append(path.read_text())

    # Ensure proper typing: first 3 are str|None, last is list[SingleTurnShot]|None
    system, user, assistant, shots = turn_templates
    return TurnTemplates(
        system=system if isinstance(system, str) else None,
        user=user if isinstance(user, str) else None,
        assistant=assistant if isinstance(assistant, str) else None,
        shots=shots if isinstance(shots, list) else None,
    )


def hash_dataset(dataset: Sequence[StringThread], num_samples: int = 5) -> str:
    """Compute a hash of dataset for quick unsafe comparison.

    Hashes: dataset length + first N elements + last N elements
    This catches most dataset changes without processing the entire dataset.

    Args:
        dataset: List of dataset items
        num_samples: Number of elements to sample from start/end (default 5)

    Returns:
        SHA256 hash
    """
    hasher = hashlib.sha256()

    hasher.update(str(len(dataset)).encode())

    num_samples = min(num_samples, len(dataset) // 2)

    for item in dataset[:num_samples]:
        hasher.update(str(item).encode())

    for item in dataset[-num_samples:]:
        hasher.update(str(item).encode())

    return hasher.hexdigest()



import html
from abc import ABC, abstractmethod
from typing import Any, ClassVar

from pybars import Compiler
from pydantic import BaseModel, Field

from adaptive_harmony import Grade, InferenceModel, StringThread, StringTurn
from adaptive_harmony.core.structured_output import JsonParseError, render_schema
from adaptive_harmony.core.utils import stringify_thread
from adaptive_harmony.graders import BaseGrader
from adaptive_harmony.graders.utils import (
    FailedJudgeLog,
    SuccessJudgeLog,
    separate_context_from_last_user_turn,
    validate_thread_last_assistant,
)

OPENAI_MODEL_FAMILIES_TEMPERATURE_1_ONLY = ["gpt-5", "o1", "o3", "o4"]


def _turns_to_dicts(turns: list[StringTurn]) -> list[dict[str, str]]:
    return [{"role": turn.role, "content": turn.content} for turn in turns]


class BaseTemplatedPromptJudgeOutput(ABC, BaseModel):
    @abstractmethod
    def get_score(self) -> float:
        pass

    def get_reasoning(self) -> str | None:
        return None


class SimpleReasonedFloatOutput(BaseTemplatedPromptJudgeOutput):
    """Sample default output format for simple float scores."""

    score: float = Field(description="The numerical score for the sample")
    reasoning: str = Field(description="Reasoning behind the score")

    def get_score(self) -> float:
        return self.score

    def get_reasoning(self) -> str:
        """Extract the reasoning from this output."""
        return self.reasoning


class BinaryJudgeOutput(BaseTemplatedPromptJudgeOutput):
    """Output format for binary PASS/FAIL/NA judges."""

    reasoning: str = Field(description="Reasoning string to support the rationale behind the score")
    score: str = Field(description="The literal score for the sample", pattern="^(PASS|FAIL|NA)$")

    SCORES_MAP: ClassVar[dict[str, float]] = {"PASS": 1.0, "FAIL": 0.0}

    def get_score(self) -> float:
        """Convert PASS/FAIL/NA to float. NA raises an exception."""
        if self.score not in self.SCORES_MAP:
            from adaptive_harmony.graders.exceptions import IgnoreScoreException

            raise IgnoreScoreException(f"Non applicable score: {self.reasoning}")
        return self.SCORES_MAP[self.score]

    def get_reasoning(self) -> str:
        return self.reasoning


class TemplatedPromptJudgeGrader[T: BaseTemplatedPromptJudgeOutput](BaseGrader[SuccessJudgeLog | FailedJudgeLog]):
    r"""
    A flexible grader that uses Handlebars templates for system and user prompts.
    Generic over T, which must inherit from BaseTemplatedPromptJudgeOutput.
    This enforces that all output models implement get_score() and get_reasoning() methods.
    Templates have access to comprehensive context extracted from the StringThread:
    - output_schema: Expected output structure schema as a string
    - turns: List of all turns as dicts with "role" and "content" keys
    - metadata: Thread metadata dict
    - context_turns: All turns without the assistant's completion (includes system prompt)
    - context_str: Context formatted as string ('turn.role:\nturn.content\n' for every turn)
    - context_turns_without_last_user: Same as context_turns, but without the last user turn
    - context_str_without_last_user: Same as context_str, but without the last user turn
    - last_user_turn_content: Content of the last user turn
    - completion: Assistant's completion
    Example Handlebars templates:
    System: "You are a judge. Evaluate responses based on: {{criteria}}.
             Always output the following JSON schema, with no preamble or postamble: {{output_schema}}"
    User: "Conversation context:\n {{context_str_without_last_user}}
           Question:\n {{last_user_turn_content}}
           Response:\n {{completion}}"
    Advanced examples:
    - Conditionals: "{{#if metadata.domain}}Domain: {{metadata.domain}}{{/if}}"
    - Loops with element index: "{{#each user_turns}}Turn {{@index}}: {{content}}{{/each}}"
    - Built-in vars: "{{@index}}, {{@key}}, {{@first}}, {{@last}}"
    """

    def __init__(
        self,
        grader_key: str,
        model: InferenceModel,
        system_template: str,
        user_template: str,
        output_model: type[T],
        template_variables: dict[str, Any] | None = None,
        temperature: float = 0.0,
        grader_id: str | None = None,
    ):
        """
        Initialize the templated prompt judge.
        Args:
            grader_key: Unique identifier for this grader
            model: Live InferenceModel to use as judge for grading
            system_template: Handlebars template string for system prompt
            user_template: Handlebars template string for user prompt
            output_model: Pydantic model class that inherits from BaseOutput
                         (must implement get_score() and get_reasoning() methods)
            template_variables: Additional variables to make available in templates
            grader_id: Optional grader ID (defaults to grader_key)
        """
        super().__init__(grader_key)
        self._logs: list[SuccessJudgeLog | FailedJudgeLog] = []  # type: ignore[assignment]

        # Set temperature to 1.0 if model_key is an OpenAI model in the temperature-1-only list
        model_path: str = model.get_builder_args().get("path")  # type: ignore[assignment]
        if model_path.startswith("openai://"):
            model_name = model_path.removeprefix("openai://").split("?")[0]
            if any(model_name.startswith(model) for model in OPENAI_MODEL_FAMILIES_TEMPERATURE_1_ONLY):
                temperature = 1.0
        self.model = model.temperature(temperature)
        self.grader_id_or_key = grader_id or grader_key

        # Template setup
        self.compiler = Compiler()
        self.system_template = self.compiler.compile(system_template)
        self.user_template = self.compiler.compile(user_template)
        self.template_variables = template_variables or {}

        # Output configuration
        self.output_model = output_model

    @classmethod
    def render_template(cls, template: str, thread: StringThread, output_model: type[BaseModel]):
        compiler = Compiler()
        compiled_template = compiler.compile(template)
        context = cls.extract_template_context(thread, output_model)
        return html.unescape(compiled_template(context))

    @classmethod
    def extract_template_context(
        cls,
        thread: StringThread,
        output_model: type[BaseModel],
        template_variables: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Extract context from StringThread for template rendering"""
        validate_thread_last_assistant(thread)

        turns = thread.get_turns()
        context_without_last_user, last_user_turn = separate_context_from_last_user_turn(
            thread, include_system_prompt=True
        )
        context = [turn for turn in turns[:-1]]

        # Build comprehensive context
        context = {
            # Core thread data
            "metadata": thread.metadata,
            "turns": _turns_to_dicts(turns),
            # Context and key turns
            "context_turns": _turns_to_dicts(context),
            "context_str": stringify_thread(StringThread(context)),
            "context_turns_without_last_user": _turns_to_dicts(context_without_last_user),
            "context_str_without_last_user": stringify_thread(StringThread(context_without_last_user)),
            "last_user_turn_content": last_user_turn,
            "completion": thread.last_content(),
            "output_schema": render_schema(output_model),
            # Additional template variables
            **(template_variables or {}),
        }

        return context

    def _build_judge_prompt(self, thread: StringThread) -> StringThread:
        """Build the judging prompt using Handlebars templates"""
        context = self.extract_template_context(thread, self.output_model, self.template_variables)

        # Render templates using pybars with helpers
        system_content = html.unescape(self.system_template(context))
        user_content = html.unescape(self.user_template(context))

        # Build prompt thread
        judge_thread = StringThread().system(system_content).user(user_content)
        return judge_thread

    async def grade(self, sample: StringThread) -> Grade:
        """Grade a sample using the templated prompt"""

        judging_prompt = self._build_judge_prompt(sample)
        str_prompt = stringify_thread(judging_prompt, sep=f"\n\n{'-' * 10}\n\n")

        try:
            _, parsed_output = await self.model.generate_and_validate(judging_prompt, self.output_model)
        except JsonParseError as e:
            self.add_log({"prompt": str_prompt, "error": f"{str(e)}\n\nCOMPLETION:\n{e.completion}"})
            raise
        except Exception as e:
            self.add_log({"prompt": str_prompt, "error": str(e)})
            raise

        # Extract score and reasoning using the abstract methods
        score = parsed_output.get_score()
        reasoning = parsed_output.get_reasoning() or ""

        grade = Grade(value=score, grader_key=self.grader_id_or_key, reasoning=reasoning)
        self.add_log({"score": score, "prompt": str_prompt, "reasoning": reasoning})

        return grade

    def get_logs(self, clear: bool = False, log_all_samples: bool = False) -> dict[str, float | Any]:
        """Get aggregated logs from all grading calls"""
        # Get base statistics
        logs = super().get_logs(clear=False)

        # Get sample logs
        successfully_scored_samples = [log for log in self._logs if "score" in log]
        failed_scored_samples = [log for log in self._logs if "error" in log]

        if not log_all_samples and successfully_scored_samples:
            # Limit samples for display
            successfully_scored_samples = successfully_scored_samples[:10]

        sample_logs = self.get_sample_tables(successfully_scored_samples, failed_scored_samples)
        logs.update(sample_logs)

        if clear:
            self.clear_logs()

        return logs



import base64
import html
import io
import os

from PIL import Image

from adaptive_harmony import StringThread


def save_inlined_string_to_png(image_str: str, path_to_save: str):
    _, base64_str = image_str.split(",")
    image_data = base64.b64decode(base64_str)
    image = Image.open(io.BytesIO(image_data))
    image.save(path_to_save)


def save_string_thread_to_markdown_dir(thread: StringThread, directory_path: str):
    all_fragments = thread.get_fragments()

    os.makedirs(directory_path, exist_ok=True)
    image_counter = 0
    with open(os.path.join(directory_path, "main.md"), "w") as w:
        for role, fragments in all_fragments:
            w.write(f"# {role}\n")
            for fragment in fragments:
                if fragment["type"] == "text":
                    w.write(fragment["text"])
                else:
                    assert fragment["type"] == "image"
                    filename = f"image_{image_counter}.jpeg"
                    image_path = os.path.join(directory_path, filename)
                    save_inlined_string_to_png(fragment["url"], image_path)
                    w.write(f"![image info]({filename})\n")


def string_thread_to_html_string(thread: StringThread) -> str:
    """
    Converts a StringThread into a single, valid HTML string.
    Text is converted to paragraphs, and images are embedded directly
    using their base64 data URIs.
    """
    html_parts = []
    all_fragments = thread.get_fragments()

    for role, fragments in all_fragments:
        # Use a heading for the role (e.g., 'user', 'model')
        html_parts.append(f"<h2>{role.capitalize()}</h2>")
        for fragment in fragments:
            if fragment["type"] == "text":
                # Escape HTML special characters. CSS will handle wrapping and newlines.
                text_content = html.escape(fragment["text"])
                html_parts.append(f"<p>{text_content}</p>")
            elif fragment["type"] == "image":
                # The 'url' is the data URI, which can be used directly in the <img> src
                image_url = fragment["url"]
                html_parts.append(
                    f'<img src="{image_url}" alt="Embedded image content" style="max-width: 500px; height: auto; display: block; margin: 10px 0;">'
                )

    # Combine all parts into a single string
    body_content = "\n".join(html_parts)

    # Wrap the content in a complete, styled HTML document
    return f"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>StringThread Conversation</title>
    <style>
        body {{
            font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Helvetica, Arial, sans-serif;
            line-height: 1.6;
            color: #000000; 
            margin: 20px auto;
            padding: 0 20px;
        }}
        h2 {{
            border-bottom: 2px solid #eee;
            padding-bottom: 10px;
            margin-top: 40px;
            margin-left: 40px;
            width: 600px;
            color: #000000; 
        }}
        p {{
            width: 600px;
            margin: 16px 0;
            margin-left: 40px;
            word-break: break-word; /* Helps with breaking long words if needed */
            white-space: pre-wrap; /* Preserve newlines and wrap long lines */
        }}
        img {{
            border: 1px solid #ddd;
            border-radius: 8px;
            padding: 5px;
            background-color: #ffffff; /* Changed to white */
        }}
    </style>
</head>
<body>
    <div style="background-color: #ffffff; width: 680px;">
        <h1>Conversation Thread</h1>
        {body_content}
    </div>
</body>
</html>
"""



import textwrap
from typing import NamedTuple, TypedDict

from pydantic import BaseModel

from adaptive_harmony import StringThread, StringTurn
from adaptive_harmony.core.structured_output import render_pydantic_model
from adaptive_harmony.core.utils import stringify_thread
from adaptive_harmony.graders.range_judge.types import PromptBuildingBlocks, ReasonedScore
from adaptive_harmony.graders.utils import (
    separate_context_from_last_user_turn,
    validate_thread_last_assistant,
)


class RangeJudgeShot(BaseModel):
    thread: StringThread
    reasoning: str
    score: int

    class Config:
        arbitrary_types_allowed = True


class RangeShots(TypedDict):
    reasoning: list[StringTurn]
    scoring: list[StringTurn]


class SubrangeExpectations(NamedTuple):
    subrange: tuple[int, int]
    expectation: str


COMMON_EVALUATION_PROMPT = """You are an expert evaluator of AI-user interactions.
You will be given:
- CONTEXT : previous conversation history, might be empty
- LAST USER INPUT : the latest input from the user
- LAST ASSISTANT OUTPUT : the latest output/answer from the AI
- CRITERIA: the evaluation criteria
- EVALUATION STEPS : logical reasoning steps to take when evaluating the interaction against the CRITERIA
"""


def get_common_user_template(
    context: str, last_user_input: str, assistant_answer: str, criteria: str, evaluation_steps: str
):
    return f"""CONTEXT\n{context}\n
LAST USER INPUT\n{last_user_input}\n
LAST ASSISTANT OUTPUT\n{assistant_answer}\n
CRITERIA\n{criteria}\n
EVALUATION STEPS\n{evaluation_steps}"""


class RangeScorerTemplates:
    @staticmethod
    def get_evaluation_steps(criteria: str) -> StringThread:
        return (
            StringThread()
            .system(
                textwrap.dedent(
                    """\
                    Given an evaluation criteria which outlines how you should judge an interaction between an AI and a user, generate 3-4 concise evaluation steps based on the criteria below.

                    Return your evaluation steps as a numbered list of evaluation steps, such as:

                    Steps list:
                    1. First step
                    2. Second step
                    3. Third step
                    etc.

                    Focus on specific, concise steps that can be objectively followed based on the evaluation criteria provided.
                    Don't return any preamble or explanation, only the list.
                    """
                )
            )
            .user(f"Evaluation Criteria:\n{criteria}\nSteps list:\n")
        )

    @staticmethod
    def get_json_reasoned_score_user(
        context: str, last_user_input: str, assistant_answer: str, criteria: str, evaluation_steps: str
    ):
        common = get_common_user_template(context, last_user_input, assistant_answer, criteria, evaluation_steps)
        return f"{common}\n\nJSON OUTPUT:\n"

    @staticmethod
    def get_json_reasoned_score(
        context: str,
        last_user_input: str,
        assistant_answer: str,
        criteria: str,
        evaluation_steps: str,
        score_range: tuple[int, int],
        json_schema: str,
        subrange_expectations: list[SubrangeExpectations] | None = None,
        shots: list[StringTurn] | None = None,
    ) -> StringThread:
        if subrange_expectations:
            subrange_expectations_str = "which should correspond to:\n" + "\n".join(
                [f"{sub.subrange[0]} - {sub.subrange[1]}: {sub.expectation}" for sub in subrange_expectations]
            )
        else:
            subrange_expectations_str = f"where {score_range[1]} indicates strong alignment with the criteria and {score_range[0]} indicates no alignment."

        system_prompt = (
            COMMON_EVALUATION_PROMPT
            + f"""
Your task is to evaluate and score the ASSISTANT ANSWER, strictly following the provided EVALUATION STEPS to evaluate the CRITERIA.
You must respond with a valid JSON object that matches the following schema:

{json_schema}

Your reasoning for the score:
- Be specific and grounded in the EVALUATION STEPS
- Uphold the evaluation objective and nuances expressed in the CRITERIA as the main target
- Mention specific details, strenghts or shortcomings of the answer, referencing relevant details from the input
- Be concise, clear, and focused on the evaluation logic.
- **Never** quote the score itself in the explanation; focus only on reasoning through the evaluation steps

Your final evaluation score must be strictly within the range of [{score_range[0]} - {score_range[1]}], {subrange_expectations_str}

Return only the JSON object after the OUTPUT header, no other text, preamble or explanation.
"""
        )

        user_prompt = RangeScorerTemplates.get_json_reasoned_score_user(
            context, last_user_input, assistant_answer, criteria, evaluation_steps
        )

        shots = shots or []
        return StringThread([("system", system_prompt)] + shots + [("user", user_prompt)])

    @staticmethod
    def get_up_to_score_user(
        context: str, last_user_input: str, assistant_answer: str, criteria: str, evaluation_steps: str, reasoning: str
    ) -> str:
        common = get_common_user_template(context, last_user_input, assistant_answer, criteria, evaluation_steps)
        return f"{common}\n\nREASONING\n{reasoning}\n\nSCORE: "

    @staticmethod
    def get_up_to_score(
        context: str,
        last_user_input: str,
        assistant_answer: str,
        criteria: str,
        evaluation_steps: str,
        score_range: tuple[int, int],
        reasoning: str,
        subrange_expectations: list[SubrangeExpectations] | None = None,
        shots: list[StringTurn] | None = None,
    ) -> StringThread:
        if subrange_expectations:
            subrange_expectations_str = "which should correspond to:\n" + "\n".join(
                [f"{sub.subrange[0]} - {sub.subrange[1]}: {sub.expectation}" for sub in subrange_expectations]
            )
        else:
            subrange_expectations_str = f"where {score_range[1]} indicates strong alignment with the evaluation steps and {score_range[0]} indicates no alignment."

        system_prompt = COMMON_EVALUATION_PROMPT + textwrap.dedent(
            f"""\
                - REASONING : the reasoning for the score, following the process described by the EVALUATION STEPS to assess the presented interaction against the CRITERIA

                You must respond only with a score, based on the original CRITERIA and the REASONING for the sample,
                which should justify your score for the sample.

                Your final evaluation score must be strictly within the range of [{score_range[0]} - {score_range[1]}], {subrange_expectations_str}

                Return only the integer score, nothing before or after.
                """
        )
        user_prompt = RangeScorerTemplates.get_up_to_score_user(
            context, last_user_input, assistant_answer, criteria, evaluation_steps, reasoning
        )

        shots = shots or []
        return StringThread([("system", system_prompt)] + shots + [("user", user_prompt)])


def get_prompt_building_blocks(thread: StringThread) -> PromptBuildingBlocks:
    validate_thread_last_assistant(thread)
    context_turns, last_user_turn = separate_context_from_last_user_turn(thread)
    context_str = stringify_thread(StringThread(context_turns))
    last_assistant_turn = thread.last_content()
    assert last_user_turn, "There must be at least one user turn"
    return PromptBuildingBlocks(
        context=context_str, last_user_turn=last_user_turn, last_assistant_turn=last_assistant_turn
    )


def create_shots(criteria: str, evaluation_steps: str, shots: list[RangeJudgeShot]) -> RangeShots:
    reasoning_shots: list[StringTurn] = []
    scoring_shots: list[StringTurn] = []
    for shot in shots:
        prompt_components = get_prompt_building_blocks(shot.thread)
        reasoning_shots.extend(
            [
                StringTurn(
                    role="user",
                    content=RangeScorerTemplates.get_json_reasoned_score_user(
                        prompt_components.context,
                        prompt_components.last_user_turn,
                        prompt_components.last_assistant_turn,
                        criteria,
                        evaluation_steps,
                    ),
                ),
                StringTurn(
                    role="assistant",
                    content=render_pydantic_model(ReasonedScore(reasoning=shot.reasoning, score=shot.score)),
                ),
            ]
        )
        scoring_shots.extend(
            [
                StringTurn(
                    role="user",
                    content=RangeScorerTemplates.get_up_to_score_user(
                        prompt_components.context,
                        prompt_components.last_user_turn,
                        prompt_components.last_assistant_turn,
                        criteria,
                        evaluation_steps,
                        shot.reasoning,
                    ),
                ),
                StringTurn(role="assistant", content=str(shot.score)),
            ]
        )

    return {"reasoning": reasoning_shots, "scoring": scoring_shots}



import os
from dataclasses import dataclass
from typing import Callable, Sequence

import numpy as np

from adaptive_harmony import (
    CombinedSchedule,
    CosineScheduler,
    DataSet,
    InferenceModel,
    JobNotifier,
    Logger,
    StageNotifier,
    StringThread,
    TokenizedThread,
    TrainingModel,
)
from adaptive_harmony.common.callbacks import RecipeCallback
from adaptive_harmony.common.checkpointing import CheckpointManager
from adaptive_harmony.core import rl_utils
from adaptive_harmony.core.utils import async_map, async_map_batch, get_minibatches, hash_hyperparams, log_args
from adaptive_harmony.graders import BaseGrader
from adaptive_harmony.metric_logger import StdoutLogger


@dataclass
class Sample:
    sample: TokenizedThread
    string_sample: StringThread
    logprobs: list[float]
    ref_logprobs: list[float]
    advantages: list[float]
    returns: list[float]
    score: float
    kl_div: list[float]
    values: list[float]
    kl_pen: float
    cumulative_reward: float


PPO_HYPERPARAMS = {
    "max_num_ppo_steps",
    "value_only_fraction",
    "lr_policy",
    "lr_scheduler_policy",
    "lr_scheduler_value",
    "lr_value",
    "samples_per_batch",
    "samples_per_mini_batch",
    "mini_epochs_per_batch",
    "max_grad_norm",
    "clip_range",
    "kl_beta",
    "gae_lambda",
    "gae_gamma",
    "weight_decay",
    "skip_nan_gradients",
}


class PPO:
    @log_args
    @hash_hyperparams(include=PPO_HYPERPARAMS)
    def __init__(
        self,
        dataset: list[StringThread],
        model: TrainingModel,
        value_model: TrainingModel,
        grader: BaseGrader,
        logger: Logger = StdoutLogger(),
        stage_notifier: StageNotifier = JobNotifier().stage_notifier("PPO Training"),
        callbacks: Sequence[RecipeCallback] = [],
        max_num_ppo_steps: int | None = None,
        value_only_fraction=0.25,
        lr_policy: float = 0.75e-6,
        lr_scheduler_policy: Callable[[float], float] | None = None,
        lr_scheduler_value: Callable[[float], float] | None = None,
        lr_value: float = 1e-6,
        samples_per_batch=128,
        samples_per_mini_batch=128,
        mini_epochs_per_batch=1,
        max_grad_norm=1.0,
        clip_range=0.1,
        kl_beta=0.1,
        gae_lambda=0.95,
        gae_gamma=1.0,
        weight_decay: float = 0,
        skip_nan_gradients: bool = False,
        restart_from_checkpoint: str | None = None,
        checkpoint_frequency: float = 0.2,
    ):
        assert value_model.is_scalar(), "You must give a scalar model to PPO for the value network"
        # Core components
        self.model_ref: InferenceModel | None = None  # Instantiated when run() is called
        self.dataset = DataSet(dataset, allow_looping=True)
        self.model = model
        self.value_model = value_model
        self.grader = grader
        self.scoring_fn = grader.score_float_value
        self.logger = logger
        self.stage_notifier = stage_notifier
        self.callbacks = callbacks
        self.skip_nan_gradients = skip_nan_gradients
        # PPO HP's
        self.max_num_batches = max_num_ppo_steps
        self.lr_schedule_policy = lr_scheduler_policy or CombinedSchedule(
            lambda _: 0, CosineScheduler(lr_policy), value_only_fraction
        )
        self.lr_schedule_value = lr_scheduler_value or CosineScheduler(lr_value)
        self.samples_per_batch = samples_per_batch
        self.samples_per_mini_batch = samples_per_mini_batch
        self.total_num_samples = (
            self.max_num_batches * self.samples_per_batch if self.max_num_batches else len(self.dataset)
        )
        self.mini_epochs_per_batch = mini_epochs_per_batch
        self.max_grad_norm = max_grad_norm
        self.clip_range = clip_range
        self.kl_beta = kl_beta
        self.gae_lambda = gae_lambda
        self.gae_gamma = gae_gamma
        self.weight_decay = weight_decay

        self.num_batches_processed = 0

        self.checkpoint_manager = CheckpointManager(
            recipe_name="PPO",
            dataset=self.dataset,
            threads_dataset=dataset,
            callbacks=self.callbacks,
            hyperparams_hash=self._hyperparams_hash,  # type: ignore
            job_id=os.environ.get("HARMONY_JOB_ID"),
            checkpoint_frequency=checkpoint_frequency,
            restart_from_checkpoint=restart_from_checkpoint,
        )

    @property
    def training_completion_percentage(self):
        return (
            self.dataset.completion_percentage()
            if self.max_num_batches is None
            else min(self.num_batches_processed / self.max_num_batches, 1.0)
        )

    async def generate_sample(self, prompt: StringThread):
        assert self.model_ref is not None, "Calling `generate_sample` before reference model has been set"

        sample = await self.model.generate_tokens(prompt)
        string_sample = await self.model.detokenize_thread(sample)
        score = await self.scoring_fn(string_sample)
        values = await self.value_model.score(sample)

        logprobs = await self.model.logprobs_per_token(sample)
        ref_logprobs = await self.model_ref.logprobs_per_token(sample)

        kl = np.array(logprobs, dtype=np.float32) - np.array(ref_logprobs, dtype=np.float32)
        kl_pen = -kl * self.kl_beta
        rewards = np.array(kl_pen)
        rewards[-1] += score

        advantages = rl_utils.gae_advantages(values, rewards.tolist(), self.gae_lambda, self.gae_gamma)
        returns = rl_utils.discounted_cumulative_rewards(rewards.tolist(), self.gae_gamma)

        return Sample(
            sample=sample,
            string_sample=string_sample,
            logprobs=logprobs,
            ref_logprobs=ref_logprobs,
            advantages=advantages,
            returns=returns,
            score=score,
            values=values,
            cumulative_reward=sum(rewards),
            kl_div=kl.tolist(),
            kl_pen=np.sum(kl_pen).item(),
        )

    async def train_ppo(self, sample: Sample):
        await self.model.train_ppo(sample.sample, sample.logprobs, sample.advantages, self.clip_range)

    async def train_value(self, sample: Sample):
        await self.value_model.train_mse_per_token(sample.sample, sample.returns)

    async def _recipe_specific_checkpoint_loading(self, checkpoint_data: dict) -> None:
        self.num_batches_processed = checkpoint_data["num_batches_processed"]

        model_checkpoint_name = checkpoint_data["model_checkpoint_name"]
        await self.model.load(f"model_registry://{model_checkpoint_name}")
        value_model_checkpoint_name = checkpoint_data["value_model_checkpoint_name"]
        await self.value_model.load(f"model_registry://{value_model_checkpoint_name}")

        self.model.set_optim_step(checkpoint_data["optim_step"])
        self.value_model.set_optim_step(checkpoint_data["value_optim_step"])

    async def _recipe_specific_checkpoint_saving(self) -> dict:
        progress_pct = int(self.training_completion_percentage * 100)
        model_checkpoint_name = f"checkpoint-{self.checkpoint_manager.job_id}-{progress_pct}-policy"
        model_checkpoint_name = await self.model.save(model_checkpoint_name, inference_only=False)

        value_model_checkpoint_name = f"checkpoint-{self.checkpoint_manager.job_id}-{progress_pct}-value"
        value_model_checkpoint_name = await self.value_model.save(value_model_checkpoint_name, inference_only=False)

        return {
            "num_batches_processed": self.num_batches_processed,
            "model_checkpoint_name": model_checkpoint_name,
            "value_model_checkpoint_name": value_model_checkpoint_name,
            "optim_step": self.model.get_optim_step(),
            "value_optim_step": self.value_model.get_optim_step(),
        }

    async def run(self):
        self.model_ref = await self.model.clone_inf()
        await self.checkpoint_manager.maybe_restore_checkpoint(self._recipe_specific_checkpoint_loading)

        self.stage_notifier.report_progress(
            tot_num_samples=self.total_num_samples,
            processed_num_samples=self.dataset.idx,
            monitoring_link=self.logger.training_monitoring_link,
        )

        while self.training_completion_percentage < 1.0:
            self.num_batches_processed += 1

            for callback in self.callbacks:
                if logs := await callback.maybe_call(self.training_completion_percentage):
                    self.logger(logs)

            # Generate training samples
            data = await async_map_batch(
                self.generate_sample,
                self.dataset,
                self.samples_per_batch,
            )
            scorer_logs = self.grader.get_logs(clear=True)
            batch_logs = {
                **{f"rewards/{key}": value for key, value in scorer_logs.items()},
                **self.get_train_batch_logs(data),
            }

            lr_policy = self.lr_schedule_policy(self.training_completion_percentage)
            lr_value = self.lr_schedule_value(self.training_completion_percentage)

            # Train on generated samples
            if lr_policy > 0:
                minibatches = get_minibatches(data, self.samples_per_mini_batch, self.mini_epochs_per_batch)
                for idx, mini_batch in enumerate(minibatches):
                    await async_map(self.train_ppo, mini_batch)
                    optim_logs = await self.model.optim_step(
                        lr_policy,
                        wd=self.weight_decay,
                        max_grad_norm=self.max_grad_norm,
                        skip_nan_gradients=self.skip_nan_gradients,
                    )
                    if idx == len(minibatches) - 1:
                        # only log tables and full batch-related logs on the final minibatch
                        self.logger(optim_logs | batch_logs)
                    else:
                        self.logger(optim_logs | dict(completion_percentage=self.training_completion_percentage))

            for mini_batch in get_minibatches(data, self.samples_per_mini_batch, self.mini_epochs_per_batch):
                await async_map(self.train_value, mini_batch)
                batch_logs |= await self.value_model.optim_step(
                    lr_value, wd=0, max_grad_norm=self.max_grad_norm, skip_nan_gradients=self.skip_nan_gradients
                )
                self.logger(batch_logs)

            self.stage_notifier.report_progress(
                tot_num_samples=self.total_num_samples,
                processed_num_samples=self.dataset.idx,
                monitoring_link=self.logger.training_monitoring_link,
            )

            if await self.checkpoint_manager.maybe_checkpoint(
                self.training_completion_percentage, self._recipe_specific_checkpoint_saving
            ):
                break

    def get_train_batch_logs(self, data: list[Sample]) -> dict:
        returns = np.concatenate([batch.returns for batch in data])
        cur_values = np.concatenate([batch.values for batch in data])

        var_return = returns.var()
        mean_error = ((cur_values - returns) ** 2).mean()
        explained_variance = (1 - mean_error / (var_return + 1e-8)).item()

        logs = dict(
            completion_percentage=self.training_completion_percentage,
            score_mean=np.mean([batch.score for batch in data]).item(),
            score_std=np.std([batch.score for batch in data]).item(),
            returns=np.mean(np.concatenate([batch.returns for batch in data])),
            kl_div=np.mean(np.concatenate([batch.kl_div for batch in data])),
            advantages=np.mean(np.concatenate([batch.advantages for batch in data])),
            generation_length=np.mean([batch.sample.len_last_turn() for batch in data]),
            logprobs=np.mean(np.concatenate([batch.logprobs for batch in data])),
            ref_logprobs=np.mean(np.concatenate([batch.ref_logprobs for batch in data])),
            kl_penalty=np.mean([batch.kl_pen for batch in data]),
            explained_variance=explained_variance,
            cumulative_reward=np.mean([batch.cumulative_reward for batch in data]),
        ) | {
            "training/completion_percentage": self.training_completion_percentage
        }  # to have an comparable axis with prior runs

        return logs



import math
from typing import Callable

Scheduler = Callable[[float], float]


class CosineSchedulerWithoutWarmup:
    def __init__(self, lr=1e-5, decay_factor=10.0) -> None:
        self.max_value = lr
        self.min_value = self.max_value / decay_factor

    def __call__(self, completion_percentage: float) -> float:
        coefficient = 0.5 * (math.cos(math.pi * completion_percentage) + 1.0)
        value_delta = self.max_value - self.min_value
        return self.min_value + coefficient * value_delta


class CombinedSchedule:
    def __init__(self, a: Scheduler, b: Scheduler, change_point: float) -> None:
        self.a = a
        self.b = b
        self.change_point = change_point

    def __call__(self, completion_percentage: float) -> float:
        if completion_percentage < self.change_point:
            return self.a(completion_percentage / self.change_point)
        else:
            return self.b((completion_percentage - self.change_point) / (1 - self.change_point))


class CosineScheduler:
    def __init__(self, lr=1e-5, warmup_percentage=0.1, decay_factor=10.0):
        self.combined = CombinedSchedule(
            lambda x: x * lr, CosineSchedulerWithoutWarmup(lr, decay_factor), warmup_percentage
        )

    def __call__(self, completion_percentage: float) -> float:
        return self.combined(completion_percentage)



from .callbacks import (
    CheckpointCallback as CheckpointCallback,
)
from .callbacks import (
    EnvironmentValidationCallback as EnvironmentValidationCallback,
)
from .callbacks import (
    GenerateSamplesCallback as GenerateSamplesCallback,
)
from .callbacks import (
    GraderEvalCallback as GraderEvalCallback,
)
from .callbacks import (
    RecipeCallback as RecipeCallback,
)
from .callbacks import (
    ValidationLossCallback as ValidationLossCallback,
)
from .dpo import DPO as DPO
from .env_grpo import ENVGRPO
from .grpo import GRPO as GRPO
from .gspo import GSPO as GSPO
from .ppo import PPO as PPO
from .rm import RewardModelling as RewardModelling
from .sft import SFT as SFT

__all__ = [
    "SFT",
    "PPO",
    "GRPO",
    "ENVGRPO",
    "DPO",
    "RewardModelling",
    "RecipeCallback",
    "GenerateSamplesCallback",
    "ValidationLossCallback",
    "CheckpointCallback",
    "GraderEvalCallback",
    "EnvironmentValidationCallback",
]



from typing import Callable, Sequence

from tqdm.auto import tqdm

from adaptive_harmony import CosineScheduler, DataSet, JobNotifier, Logger, StageNotifier, StringThread, TrainingModel
from adaptive_harmony.common.callbacks import RecipeCallback
from adaptive_harmony.core.utils import async_map_batch, log_args
from adaptive_harmony.metric_logger import StdoutLogger


class RewardModelling:
    @log_args
    def __init__(
        self,
        dataset: list[tuple[StringThread, StringThread]] | list[StringThread],
        model: TrainingModel,
        logger: Logger = StdoutLogger(),
        job_notifier: StageNotifier = JobNotifier().stage_notifier("Reward Model Training"),
        callbacks: Sequence[RecipeCallback] = [],
        lr: float = 1e-06,
        lr_scheduler: Callable[[float], float] | None = None,
        samples_per_batch: int = 64,
        max_grad_norm: float = 1.0,
        epochs: int = 1,
        skip_nan_gradients: bool = False,
    ):
        self.dataset: DataSet[StringThread | tuple[StringThread, StringThread]] = DataSet(dataset, allow_looping=True)
        self.lr_schedule = lr_scheduler or CosineScheduler(lr)
        self.model = model
        self.logger = logger
        self.job_notifier = job_notifier
        self.callbacks = callbacks
        self.samples_per_batch = samples_per_batch
        self.max_grad_norm = max_grad_norm
        self.skip_nan_gradients = skip_nan_gradients
        self.epochs = epochs

    @property
    def training_completion_percentage(self):
        return self.dataset.completion_percentage() / self.epochs

    async def train_rm(self, sample: tuple[StringThread, StringThread] | StringThread):
        # having both preference and metric feedback in dataset likely will never happen, but it's possible
        if isinstance(sample, tuple):
            await self.model.train_ranking(sample[0], sample[1])
        else:
            if "res" not in sample.metadata:
                raise ValueError(f"Sample missing required 'res' field in metadata: {sample.metadata.keys()}")
            target_value = sample.metadata["res"]
            await self.model.train_mse(sample, target_value)

    async def run(self):
        with tqdm(total=100) as pbar:
            self.job_notifier.report_progress(
                tot_num_samples=len(self.dataset) * self.epochs,
                processed_num_samples=self.dataset.idx,
                monitoring_link=self.logger.training_monitoring_link,
            )
            while self.training_completion_percentage < 1.0:
                for callback in self.callbacks:
                    if logs := await callback.maybe_call(self.training_completion_percentage):
                        self.logger(logs)

                await async_map_batch(self.train_rm, self.dataset, self.samples_per_batch)
                cp = self.training_completion_percentage
                current_lr = self.lr_schedule(cp)
                pbar.update(cp * 100.0 - pbar.n)

                logs = await self.model.optim_step(
                    current_lr, wd=0, max_grad_norm=self.max_grad_norm, skip_nan_gradients=self.skip_nan_gradients
                )

                self.job_notifier.report_progress(
                    tot_num_samples=len(self.dataset) * self.epochs,
                    processed_num_samples=self.dataset.idx,
                    monitoring_link=self.logger.training_monitoring_link,
                )

                self.logger(logs | dict(completion_percentage=cp))



from typing import Callable, Sequence

import numpy as np
from datasets import load_dataset

from adaptive_harmony import StringThread


class DataSet[T]:
    def __init__(self, threads: Sequence[T], allow_looping: bool = False, seed: int = 42):
        self.threads = threads
        self.allow_looping = allow_looping
        # This will be used to shuffle the dataset when we cross the epoch boundary, initially it is just the indices of the threads
        # to respect the given order in the first epoch
        self.rng = np.random.default_rng(seed)
        self.access_indices = self.rng.permutation(len(threads))
        self.idx = 0

    def __iter__(self) -> "DataSet":
        return self

    def __len__(self) -> int:
        return len(self.threads)

    def __next__(self) -> T:
        if not self.allow_looping and self.idx == len(self.threads):
            raise StopIteration()
        elif self.allow_looping and self.idx == len(self.access_indices):
            self.access_indices = np.concatenate([self.access_indices, self.rng.permutation(len(self.threads))])

        sample_idx = self.access_indices[self.idx]
        ret = self.threads[sample_idx]

        if hasattr(ret, "metadata") and isinstance(ret.metadata, dict):
            ret.metadata["sample_index"] = self.idx

        self.idx += 1

        return ret

    def __getitem__(self, x):
        return self.threads.__getitem__(x)

    def completion_percentage(self) -> float:
        """If dataset is looping, this can return a value greater than 1.0. Handle in recipe."""
        return self.idx / len(self.threads)

    def reset(self):
        self.idx = 0


def convert_sample_dict(
    turns_key: str | None = "messages", role_key="role", content_key="content", trim_final_assistant_turns=False
):
    def f(dialogue: dict) -> StringThread:
        if turns_key is not None:
            dialogue = dialogue[turns_key]
        turns = [(turn[role_key], turn[content_key]) for turn in dialogue]

        if trim_final_assistant_turns:
            while len(turns) > 0 and turns[-1][0] == "assistant":
                turns = turns[:-1]

        return StringThread(turns)

    return f


def load_from_hf(repo: str, split: str, convert_sample_fn: Callable[..., StringThread]) -> list[StringThread]:
    dataset = load_dataset(repo, split=split, keep_in_memory=True)
    dataset = dataset.select(range(len(dataset)))  # type: ignore
    return [convert_sample_fn(x) for x in dataset]



import statistics
from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, cast

from harmony_client import InferenceModel
from harmony_client.runtime.data import (
    AdaptiveGrader,
    CustomJudge,
    PrebuiltConfigKey,
    PrebuiltJudge,
    RemoteRewardEndpoint,
)

from adaptive_harmony import Grade, StringThread
from adaptive_harmony.graders.utils import FailedJudgeLog, SuccessJudgeLog
from adaptive_harmony.logging_table import Table
from adaptive_harmony.parameters import Model
from adaptive_harmony.runtime import RecipeContext


class BaseGrader[LogType](ABC):
    """
    Base Grader to inherit from when building a scoring function.
    """

    def __init__(self, grader_key: str):
        self._logs: list[LogType] = []
        self.grader_key = grader_key

    @abstractmethod
    async def grade(self, sample: StringThread) -> Grade:
        """
        Grade a single sample.
        Returns a single float score, with optional metadata.
        Metadata can be useful for evals when LLM reasoning regarding the score is available.
        """
        pass

    async def score_float_value(self, sample: StringThread) -> float:
        """Returns only the float score from .score"""
        return (await self.grade(sample)).value

    def add_log(self, log_data: LogType) -> None:
        """Add a log entry to the scorer's log collection."""
        self._logs.append(log_data)

    def get_logs(self, clear: bool = False, log_all_samples: bool = False) -> dict[str, float | Table]:
        """
        Get aggregated logs from all score calls.
        Base implementation computes statistics for "score" keys in individual logs.
        If there are none, returns empty dict.
        """
        if not self._logs:
            return {}

        scores = [s for s in [cast(dict[str, Any], log).get("score") for log in self._logs] if s is not None]
        logs = {}
        if scores:
            logs.update(
                dict(
                    **{
                        f"score/{key}": value
                        for key, value in dict(
                            mean=statistics.mean(scores),
                            std=statistics.stdev(scores) if len(scores) > 1 else 0.0,
                            min=min(scores),
                            max=max(scores),
                            count=len(scores),
                        ).items()
                    },
                )
            )
        if clear:
            self.clear_logs()
        return logs

    def clear_logs(self) -> None:
        """
        Clear all accumulated logs.
        """
        self._logs.clear()

    def get_sample_tables(
        self, successful_samples: list[SuccessJudgeLog], failed_samples: list[FailedJudgeLog] | None = None
    ):
        table_logs = {}
        scored_samples = (
            Table()
            .add_column("Prompt", [log["prompt"] for log in successful_samples])
            .add_column("Reasoning", [log.get("reasoning") for log in successful_samples])
            .add_column("Score", [float(log["score"]) for log in successful_samples])
        )
        if failed_samples:
            unscored_samples = (
                Table()
                .add_column("Prompt", [log.get("prompt") for log in failed_samples])
                .add_column("Error", [str(log["error"]) for log in failed_samples])
            )
            table_logs["score/unscored_samples"] = unscored_samples
        table_logs["score/scored_samples"] = scored_samples
        table_logs["score/unscored_samples_count"] = len(failed_samples) if failed_samples else 0
        table_logs["score/scored_samples_count"] = len(successful_samples)
        return table_logs

    @classmethod
    def from_function(
        cls, grader_key: str, async_fn: Callable[[StringThread], Awaitable[float]]
    ) -> "BaseGrader[dict[str, Any]]":
        class FunctionScorer(BaseGrader[dict[str, float]]):
            def __init__(self):
                super().__init__(grader_key)

            async def grade(self, sample: StringThread) -> Grade:
                result = await async_fn(sample)
                grade = Grade(value=result, grader_key=self.grader_key)
                self.add_log({"score": result})
                return grade

        return FunctionScorer()

    @classmethod
    async def from_config(
        cls,
        grader_config: AdaptiveGrader,
        ctx: RecipeContext,
        tp: int | None = None,
        kv_cache_len: int | None = None,
        max_tokens: int | None = None,
    ) -> "BaseGrader[dict[str, Any]]":
        match grader_config.config.type:
            case "Judge":
                config = cast(CustomJudge, grader_config.config)
                return await cls.from_templated_judge(
                    grader_config.key, str(grader_config.grader_id), config, ctx, tp, kv_cache_len, max_tokens
                )
            case "Prebuilt":
                config = cast(PrebuiltJudge, grader_config.config)
                return await cls.from_prebuilt_judge(
                    grader_config.key, str(grader_config.grader_id), config, ctx, tp, kv_cache_len
                )
            case "Remote":
                config = cast(RemoteRewardEndpoint, grader_config.config)
                return cls.from_remote_reward_endpoint(grader_config.key, str(grader_config.grader_id), config)
            case _:
                raise ValueError(f"Invalid grader type: {grader_config.config.type}")

    @classmethod
    async def from_templated_judge(
        cls,
        grader_key: str,
        grader_id: str,
        config: CustomJudge,
        ctx: RecipeContext,
        tp: int | None = None,
        kv_cache_len: int | None = None,
        max_tokens: int | None = None,
    ) -> "BaseGrader[dict[str, Any]]":
        # Import here to avoid circular dependency
        from adaptive_harmony.graders.templated_prompt_judge import (
            BinaryJudgeOutput,
            TemplatedPromptJudgeGrader,
        )

        # Convert examples to template variables
        examples = []
        for example in config.examples:
            examples.append(
                {
                    "context_str": (
                        "\n".join(f"{msg.role}:\n{msg.content}" for msg in example.input[:-1])
                        if len(example.input) > 1
                        else ""
                    ),
                    "user_question": example.input[-1].content if example.input else "",
                    "completion": example.output,
                    "output_json": f'{{"reasoning": "{example.reasoning or ""}", "score": "{"PASS" if example.pass_ else "FAIL"}"}}',
                }
            )

        template_vars = {
            "criteria": config.criteria,
            "examples": examples,
        }

        model = await get_model(ctx, grader_key, config.model_key, kv_cache_len, max_tokens, tp)

        return TemplatedPromptJudgeGrader(
            grader_key=grader_key,
            model=model,
            grader_id=grader_id,
            system_template=config.system_template,
            user_template=config.user_template,
            output_model=BinaryJudgeOutput,
            template_variables=template_vars,
        )  # type: ignore[return-value]

    @classmethod
    async def from_prebuilt_judge(
        cls,
        grader_key: str,
        grader_id: str,
        config: PrebuiltJudge,
        ctx: RecipeContext,
        tp: int | None = None,
        kv_cache_len: int | None = None,
        max_tokens: int | None = None,
    ) -> "BaseGrader[dict[str, Any]]":
        model = await get_model(ctx, grader_key, config.model_key, kv_cache_len, max_tokens, tp)

        match config.prebuilt_config_key:
            case PrebuiltConfigKey.Faithfulness:
                # Import here to avoid circular dependency
                from adaptive_harmony.graders.faithfulness_judge.faithfulness_judge import FaithfulnessGrader

                return FaithfulnessGrader(
                    grader_key=grader_key,
                    grader_id=grader_id,
                    model=model,
                )
            case PrebuiltConfigKey.AnswerRelevancy:
                # Import here to avoid circular dependency
                from adaptive_harmony.graders.answer_relevancy_judge.answer_relevancy_judge import AnswerRelevancyGrader

                return AnswerRelevancyGrader(
                    grader_key=grader_key,
                    grader_id=grader_id,
                    model=model,
                )
            case PrebuiltConfigKey.ContextRelevancy:
                # Import here to avoid circular dependency
                from adaptive_harmony.graders.context_relevancy_judge.context_relevancy_judge import (
                    ContextRelevancyGrader,
                )

                return ContextRelevancyGrader(
                    grader_key=grader_key,
                    grader_id=grader_id,
                    model=model,
                )
            case _:
                raise ValueError(f"Invalid prebuilt judge type: {config.prebuilt_config_key}")

    @classmethod
    def from_remote_reward_endpoint(
        cls, grader_key: str, grader_id: str, config: RemoteRewardEndpoint
    ) -> "BaseGrader[dict[str, Any]]":
        # Import here to avoid circular dependency
        from adaptive_harmony.graders.reward_server_grader import RewardServerGrader

        return RewardServerGrader(grader_key=grader_key, grader_id=grader_id, reward_server_ip=config.url)


async def get_model(
    ctx: RecipeContext,
    grader_key: str,
    model_key: str,
    kv_cache_len: int | None,
    max_tokens: int | None,
    tp: int | None,
) -> InferenceModel:
    model_builder = await Model(model_key=model_key).to_builder(
        ctx, kv_cache_len=kv_cache_len, tokens_to_generate=max_tokens, tp=tp
    )
    model = await model_builder.spawn_inference(grader_key)
    return model



# ruff: noqa: F403, F401
from typing import TYPE_CHECKING

from harmony_client import (
    EvalSample as EvalSample,
)
from harmony_client import (
    EvalSampleInteraction as EvalSampleInteraction,
)
from harmony_client import (
    Grade as Grade,
)
from harmony_client import (
    HarmonyClient as HarmonyClient,
)
from harmony_client import (
    HarmonyJobNotifier as HarmonyJobNotifier,
)
from harmony_client import (
    InferenceModel as InferenceModel,
)
from harmony_client import (
    JobArtifact as JobArtifact,
)
from harmony_client import (
    JobNotifier as JobNotifier,
)
from harmony_client import (
    ModelBuilder as ModelBuilder,
)
from harmony_client import (
    StageNotifier as StageNotifier,
)
from harmony_client import (
    StringThread as StringThread,
)
from harmony_client import (
    TokenizedThread as TokenizedThread,
)
from harmony_client import (
    TrainingModel as TrainingModel,
)
from harmony_client import (
    get_client as get_client,
)
from harmony_client import parameters as parameters
from harmony_client import runtime as runtime
from rich.progress import Progress

if TYPE_CHECKING:
    from harmony_client import StringTurn as StringTurn
else:
    from typing import NamedTuple

    class StringTurn(NamedTuple):
        role: str
        content: str


from harmony_client.artifacts.custom_artifact import CustomArtifact
from harmony_client.artifacts.dataset_artifact import DatasetArtifact
from harmony_client.file_storage import (
    FileStorage,
    FileStorageConfig,
    LocalFileStorageConfig,
    S3FileStorageConfig,
    StoredFile,
)

import adaptive_harmony.core.rl_utils as rl_utils
from adaptive_harmony.core.dataset import DataSet
from adaptive_harmony.core.schedulers import CombinedSchedule, CosineScheduler, CosineSchedulerWithoutWarmup, Scheduler
from adaptive_harmony.evaluation.evaluation_artifact import EvaluationArtifact
from adaptive_harmony.metric_logger import Logger, WandbLogger

# Ensure key classes are available at module level
__all__ = [
    "StringThread",
    "StringTurn",
    "TokenizedThread",
    "InferenceModel",
    "ModelBuilder",
    "TrainingModel",
    "HarmonyClient",
    "get_client",
    "DataSet",
    "CosineScheduler",
    "CombinedSchedule",
    "CosineSchedulerWithoutWarmup",
    "Scheduler",
    "WandbLogger",
    "Logger",
    "FileStorage",
    "FileStorageConfig",
    "LocalFileStorageConfig",
    "S3FileStorageConfig",
    "StoredFile",
    "EvaluationArtifact",
    "CustomArtifact",
    "DatasetArtifact",
    "rl_utils",
    "Grade",
    "EvalSample",
    "EvalSampleInteraction",
    "JobArtifact",
]


# Patch StringThread to use rich for display
from harmony_client.runtime.model_artifact_save import save_with_artifact

from adaptive_harmony.core.display import _stringthread_repr, _tokenizedthread_repr
from adaptive_harmony.core.image_utils import string_thread_to_html_string

# Patch InferenceModel to have json output capabilities
from adaptive_harmony.core.structured_output import generate_and_validate, render_pydantic_model, render_schema

StringThread.__repr__ = _stringthread_repr  # type: ignore
TokenizedThread.__repr__ = _tokenizedthread_repr  # type: ignore
setattr(StringThread, "_repr_html_", string_thread_to_html_string)
setattr(InferenceModel, "generate_and_validate", generate_and_validate)
setattr(InferenceModel, "render_schema", staticmethod(render_schema))
setattr(InferenceModel, "render_pydantic_model", staticmethod(render_pydantic_model))

_original_training_model_save = TrainingModel.save


async def _save_with_artifact_wrapper(model: TrainingModel, model_name: str, inference_only: bool = True, ctx=None):
    return await save_with_artifact(model, model_name, inference_only, ctx, _original_training_model_save)


setattr(TrainingModel, "save", _save_with_artifact_wrapper)


async def spawn_train(self: ModelBuilder, name: str, max_batch_size: int) -> TrainingModel:
    fut = await self.spawn_train_with_progress(name, max_batch_size)  # type:ignore

    with Progress() as pbar:
        task = pbar.add_task("Loading model", total=1000)

        while (prog := await fut._await_progress()) != 1.0:
            pbar.update(task, completed=prog, total=1.0)
        pbar.update(task, completed=1.0, total=1.0)

    return await fut.get()


async def spawn_inference(self: ModelBuilder, name: str) -> InferenceModel:
    fut = await self.spawn_inference_with_progress(name)  # type:ignore

    with Progress() as pbar:
        task = pbar.add_task("Loading model", total=1000)

        while (prog := await fut._await_progress()) != 1.0:
            pbar.update(task, completed=prog, total=1.0)
        pbar.update(task, completed=1.0, total=1.0)

    return await fut.get()


setattr(ModelBuilder, "spawn_inference", spawn_inference)
setattr(ModelBuilder, "spawn_train", spawn_train)



def gae_advantages(
    values: list[float],
    rewards: list[float],
    gae_lambda: float,
    gae_gamma: float,
) -> list[float]:
    response_length = len(values)

    lastgaelam = 0.0
    advantages_reversed: list[float] = []
    for t in reversed(range(response_length)):
        nextvalues = values[t + 1] if t < response_length - 1 else 0.0
        delta = rewards[t] + gae_gamma * nextvalues - values[t]
        lastgaelam = delta + gae_gamma * gae_lambda * lastgaelam
        advantages_reversed.append(lastgaelam)

    return advantages_reversed[::-1]


def discounted_cumulative_rewards(
    rewards: list[float],
    gamma: float,
) -> list[float]:
    n = len(rewards)
    returns = [0.0] * n
    returns[-1] = rewards[-1]
    for t in reversed(range(n - 1)):
        returns[t] = rewards[t] + gamma * returns[t + 1]

    return returns


def gae_td_returns(
    advantages: list[float],
    values: list[float],
) -> list[float]:
    returns = [a + b for a, b in zip(advantages, values)]
    return returns



from __future__ import annotations

import asyncio
from typing import Literal

import numpy as np
import pysbd
from harmony_client import Grade, InferenceModel, StringThread
from pydantic import BaseModel, Field

from adaptive_harmony.core.utils import stringify_thread
from adaptive_harmony.graders.base_grader import BaseGrader
from adaptive_harmony.graders.context_relevancy_judge.prompts import DEFAULT_SHOTS, SYSTEM, USER
from adaptive_harmony.graders.faithfulness_judge.faithfulness_judge import SupportedLanguages
from adaptive_harmony.graders.utils import sample_score_distribution
from adaptive_harmony.logging_table import Table


class DocumentRelevancyResult(BaseModel):
    reason: str = Field(description="The justification for the score given to a document. Keep it short and concise.")
    score: Literal[0, 1] = Field(
        description="The score for the document. A score of 1 if the document contains information relevant to answering the user input, and 0 if the document does not contain information relevant to answering the user input"
    )


class ContextRelevancyGrader(BaseGrader):
    def __init__(
        self,
        model: InferenceModel,
        language: SupportedLanguages = "en",
        grader_key: str = "context_relevancy_judge",
        grader_id: str | None = None,
    ):
        super().__init__(grader_key)
        self.model = model
        self.language = language
        self.grader_id_or_key = grader_id or grader_key
        self.sentence_splitter = pysbd.Segmenter(language=language)
        self.shots = DEFAULT_SHOTS

    async def grade(self, sample: StringThread) -> Grade:
        documents = sample.metadata.get("documents", []) if sample.metadata else []
        if not documents:
            self.add_log(
                {
                    "prompt": stringify_thread(sample, sep=f"\n\n{'-' * 10}\n\n"),
                    "error": "No document turns found in thread",
                }
            )
            raise ValueError("No document turns found in thread")

        user_question = next((turn[1] for turn in reversed(sample.get_turns()) if turn[0] == "user"), None)
        if not user_question:
            self.add_log(
                {"prompt": stringify_thread(sample, sep=f"\n\n{'-' * 10}\n\n"), "error": "No user turn found in thread"}
            )
            raise ValueError("No user turn found in thread")

        judging_threads = [
            (
                StringThread()
                .system(SYSTEM.format(json_schema=self.model.render_schema(DocumentRelevancyResult), shots=self.shots))
                .user(USER.format(user_question=user_question, document=document))
            )
            for document in documents
        ]

        try:
            judge_tasks = [
                self.model.temperature(0.0).generate_and_validate(thread, DocumentRelevancyResult)
                for thread in judging_threads
            ]
            results = await asyncio.gather(*judge_tasks)
        except Exception as e:
            self.add_log(
                {
                    "error": str(e),
                    "number_of_documents": len(documents),
                    "documents": documents,
                    "prompt": stringify_thread(judging_threads[0]),
                }
            )
            raise

        doc_relevancy_results = [result[1] for result in results]

        reason = ""
        for i, (document, doc_result) in enumerate(zip(documents, doc_relevancy_results)):
            emoji = "✅" if doc_result.score == 1 else "❌"
            result = "PASS" if doc_result.score == 1 else "FAIL"
            doc_display = document[:150] + ("..." if len(document) > 150 else "")
            reason += f"{emoji} Document {i}: {result}\n Content: {doc_display}:\nReason: {doc_result.reason}\n\n"

        score = np.mean([float(verdict.score) for verdict in doc_relevancy_results]) if doc_relevancy_results else 0.0
        self.add_log(
            {
                "score": score,
                "reasoning": reason,
                "number_of_documents": len(documents),
                "documents": documents,
                "prompt": stringify_thread(judging_threads[0]),
            }
        )
        return Grade(value=float(score), grader_key=self.grader_id_or_key, reasoning=reason)

    def get_logs(self, clear: bool = False, log_all_samples: bool = False) -> dict[str, float | Table]:
        # Only clear logs at the end if clear is True
        logs = super().get_logs(clear=False)

        successfully_scored_samples = [log for log in self._logs if "score" in log]

        # stratified sample range of scores to see high and low
        if not log_all_samples:
            subset_successfully_scored_samples = sample_score_distribution(successfully_scored_samples, 15)
        else:
            # if we have fewer than 15 samples or we want to log all samples, take them all
            subset_successfully_scored_samples = successfully_scored_samples

        failed_scored_samples = [log for log in self._logs if "error" in log]

        sample_logs = self.get_sample_tables(subset_successfully_scored_samples, failed_scored_samples)

        logs.update(sample_logs)

        if clear:
            self.clear_logs()

        return logs



SYSTEM = """You are an expert data reviewer.
You will be given a list of statements.
You task is to determine whether each statement is relevant to addressing the user's input.

Since you are going to generate a verdict for each statement, the number of `verdicts` SHOULD BE STRICTLY EQUAL to the number of `statements`, and verdicts should be in the same order as the original statements.

You always output a JSON object with the following schema, and nothing else before or after:
{json_schema}

Examples:
{shots}
"""

USER = """Your real task:
INPUT
{user_question}

STATEMENTS
{statements}

```json"""


DEFAULT_SHOTS = """INPUT
What percentage is considered a good rental yield?

STATEMENTS
0: How are you doing today?
1: Rental yield is how much you could expect to receive in rent each year from your buy to let investment.
2: Rental yield is expressed as a percentage - reflecting your rental income against the property's market value.
3: Anything around the 5-6% mark could be considered a good rental yield.
4: Anything above 6% could be considered a very good rental yield.

```json
{
    "verdicts": [
        {
            "reason": "The statement is unrelated to the input.",
            "score": 0
        },
        {
            "reason": "While the statement discusses rental yields, it does not indicate what constitutes a good rental yield.",
            "score": 0
        },
        {
            "reason": "While the statement mentions that yield is expressed as a percentage, it does not address the user question.",
            "score": 0
        },
        {
            "reason": "The statement addresses the user input, specifying what a good rental yield is.",
            "score": 1
        },
        {
            "reason": "The statement addresses the user input, specifying what a very good rental yield is.",
            "score": 1
        },
    ]
}```"""



from adaptive_harmony.core.reward_client.client import RewardClient
from adaptive_harmony.core.reward_client.reward_types import (
    MetadataValidationResponse,
    Request,
    Response,
    ServerInfo,
    Turn,
)
from adaptive_harmony.core.reward_client.websocket_utils import ResponseAccumulator

__all__ = [
    "RewardClient",
    "Turn",
    "Request",
    "Response",
    "MetadataValidationResponse",
    "ServerInfo",
    "ResponseAccumulator",
]



from __future__ import annotations

from typing import Literal

import numpy as np
import pysbd
from harmony_client import Grade, InferenceModel, StringThread
from pydantic import BaseModel, Field

from adaptive_harmony.core.utils import stringify_thread
from adaptive_harmony.graders.answer_relevancy_judge.prompts import DEFAULT_SHOTS, SYSTEM, USER
from adaptive_harmony.graders.base_grader import BaseGrader
from adaptive_harmony.graders.faithfulness_judge.faithfulness_judge import SupportedLanguages
from adaptive_harmony.graders.utils import sample_score_distribution, separate_context_from_last_user_turn
from adaptive_harmony.logging_table import Table


class StatementRelevancy(BaseModel):
    reason: str = Field(description="The justification for the score given to a statement. Keep it short and concise.")
    score: Literal[0, 1] = Field(
        description="The score for the statement. A score of 1 if the statement is relevant to addressing the original input, and 0 if the statement is irrelevant to addressing the input"
    )


class AnswerRelevancyResults(BaseModel):
    results: list[StatementRelevancy] = Field(description="A list of relevancy results for the statements")


class AnswerRelevancyGrader(BaseGrader):
    def __init__(
        self,
        model: InferenceModel,
        language: SupportedLanguages = "en",
        grader_key: str = "answer_relevancy_judge",
        grader_id: str | None = None,
    ):
        super().__init__(grader_key)
        self.model = model
        self.language = language
        self.grader_id_or_key = grader_id or grader_key
        self.sentence_splitter = pysbd.Segmenter(language=language)
        self.shots = DEFAULT_SHOTS

    async def grade(self, sample: StringThread) -> Grade:
        _, user_question = separate_context_from_last_user_turn(sample)

        completion = sample.last_content()
        split_sentences = self.sentence_splitter.segment(completion)
        sentences = [sentence.strip() for sentence in split_sentences if sentence.strip()]
        sentences_judge_str = "\n".join(f"{i}: {sentence}" for i, sentence in enumerate(sentences))

        judging_thread = (
            StringThread()
            .system(SYSTEM.format(json_schema=self.model.render_schema(AnswerRelevancyResults), shots=self.shots))
            .user(USER.format(user_question=user_question, statements=sentences_judge_str))
        )

        try:
            _, response = await self.model.temperature(0.0).generate_and_validate(
                judging_thread, AnswerRelevancyResults
            )
            results = response.results
        except Exception as e:
            self.add_log({"prompt": stringify_thread(judging_thread, sep=f"\n\n{'-' * 10}\n\n"), "error": str(e)})
            raise

        reason = ""
        for i, (result, statement) in enumerate(zip(results, sentences)):
            emoji = "✅" if result.score == 1 else "❌"
            result = "PASS" if result.score == 1 else "FAIL"
            statement_display = statement[:150] + ("..." if len(statement) > 150 else "")
            reason += f"{emoji} Statement {i}: {result}\n Excerpt: {statement_display}:\nReason: {result}\n\n"

        score = np.mean([float(result.score) for result in results]) if results else 0.0
        self.add_log(
            {"score": score, "prompt": stringify_thread(judging_thread, sep=f"\n\n{'-' * 10}\n\n"), "reasoning": reason}
        )
        return Grade(value=float(score), grader_key=self.grader_id_or_key, reasoning=reason)

    def get_logs(self, clear: bool = False, log_all_samples: bool = False) -> dict[str, float | Table]:
        # Only clear logs at the end if clear is True
        logs = super().get_logs(clear=False)

        successfully_scored_samples = [log for log in self._logs if "score" in log]

        # stratified sample range of scores to see high and low
        if not log_all_samples:
            subset_successfully_scored_samples = sample_score_distribution(successfully_scored_samples, 15)
        else:
            # if we have fewer than 15 samples or we want to log all samples, take them all
            subset_successfully_scored_samples = successfully_scored_samples

        failed_scored_samples = [log for log in self._logs if "error" in log]

        sample_logs = self.get_sample_tables(subset_successfully_scored_samples, failed_scored_samples)

        logs.update(sample_logs)

        if clear:
            self.clear_logs()

        return logs



from typing import Literal

import pysbd
from pydantic import BaseModel, Field

from adaptive_harmony import Grade, InferenceModel, StringThread
from adaptive_harmony.core.structured_output import JsonParseError
from adaptive_harmony.core.utils import stringify_thread
from adaptive_harmony.graders.base_grader import BaseGrader
from adaptive_harmony.graders.exceptions import IgnoreScoreException
from adaptive_harmony.graders.faithfulness_judge.prompts import SYSTEM, USER
from adaptive_harmony.graders.utils import (
    FailedJudgeLog,
    SuccessJudgeLog,
    sample_score_distribution,
    separate_context_from_last_user_turn,
    validate_thread_last_assistant,
)
from adaptive_harmony.logging_table import Table


class SingleStatementFaithfulnessJudgeOutput(BaseModel):
    statement_idx: int = Field(description="The original index of the sentence being scored")
    reasoning: str = Field(description="Reasoning to support the rationale behind the score")
    score: Literal["1", "0"] = Field(
        description="The score of the sample, 1 if the statement is fully supported by the context, 0 if it is not"
    )


class FaithfulnessGraderOutput(BaseModel):
    all_statements_scoring: list[SingleStatementFaithfulnessJudgeOutput] = Field(
        description="An array of objects, each analyzing a single statement from the original list of statements"
    )


SupportedLanguages = Literal[
    "en",
    "hi",
    "mr",
    "zh",
    "es",
    "am",
    "ar",
    "hy",
    "bg",
    "ur",
    "ru",
    "pl",
    "fa",
    "nl",
    "da",
    "fr",
    "my",
    "el",
    "it",
    "ja",
    "de",
    "kk",
    "sk",
]


class FaithfulnessGrader(BaseGrader):
    """
    Scores each sentence in the last assistant turn as fully supported by the context or not (1 or 0).
    The context is the rest of the thread, excluding the system prompt.
    The final score is the average of each sentence.
    Requires an input language code to split the sentences.
    """

    def __init__(
        self,
        model: InferenceModel,
        language: SupportedLanguages = "en",
        grader_key: str = "faithfulness_judge",
        grader_id: str | None = None,
    ):
        super().__init__(grader_key)
        self._logs: list[SuccessJudgeLog | FailedJudgeLog] = []  # already created in super, this is for typing
        self.model = model
        self.language = language
        self.sentence_splitter = pysbd.Segmenter(language=language)
        self.grader_id_or_key = grader_id or grader_key

    async def grade(self, sample: StringThread) -> Grade:
        # Split response into sentences
        validate_thread_last_assistant(sample)
        # Separate conversation context from last user turn
        context_turns, user_question = separate_context_from_last_user_turn(sample)
        completion = sample.last_content()
        split_sentences = self.sentence_splitter.segment(completion)
        sentences = [f"{i}: {sentence.strip()}" for i, sentence in enumerate(split_sentences) if sentence.strip()]
        sentences_judge_str = "\n".join(sentences)

        # Build prompt
        context_str = stringify_thread(StringThread(context_turns))
        judge_thread = (
            StringThread()
            .system(SYSTEM.format(json_schema=self.model.render_schema(FaithfulnessGraderOutput)))
            .user(USER.format(context=context_str, user_question=user_question, sentences=sentences_judge_str))
        )
        judge_str_prompt = stringify_thread(judge_thread, sep=f"\n\n{'-' * 10}\n\n")
        # Generate response
        try:
            _, parsed_response = await self.model.temperature(0.0).generate_and_validate(
                judge_thread, FaithfulnessGraderOutput
            )
        except JsonParseError as e:
            self.add_log({"prompt": judge_str_prompt, "error": f"{str(e)}\n\nCOMPLETION:\n{e.completion}"})
            raise
        except Exception as e:
            self.add_log({"prompt": judge_str_prompt, "error": str(e)})
            raise

        # Raise error if judge failed to judge any sentence
        n_judged_sentences = len(parsed_response.all_statements_scoring)
        if n_judged_sentences != len(sentences):
            raise IgnoreScoreException(
                f"Number of sentences in the response ({n_judged_sentences})"
                f"does not match the number of sentences in the input ({len(sentences)})"
            )
        # Calculate avg score
        score = round(
            sum([float(judgement.score) for judgement in parsed_response.all_statements_scoring]) / n_judged_sentences,
            3,
        )
        merged_reasoning_traces = "\n-".join(
            [judgement.reasoning for judgement in parsed_response.all_statements_scoring]
        )
        self.add_log({"score": score, "prompt": judge_str_prompt, "reasoning": merged_reasoning_traces})

        return Grade(value=score, grader_key=self.grader_id_or_key, reasoning=merged_reasoning_traces)

    def add_log(self, log: SuccessJudgeLog | FailedJudgeLog) -> None:
        self._logs.append(log)

    def get_logs(self, clear: bool = False, log_all_samples: bool = False) -> dict[str, float | Table]:
        # Only clear logs at the end if clear is True
        logs = super().get_logs(clear=False)

        successfully_scored_samples = [log for log in self._logs if "score" in log]

        # stratified sample range of scores to see high and low
        if not log_all_samples:
            subset_successfully_scored_samples = sample_score_distribution(successfully_scored_samples, 15)
        else:
            # if we have fewer than 15 samples or we want to log all samples, take them all
            subset_successfully_scored_samples = successfully_scored_samples

        failed_scored_samples = [log for log in self._logs if "error" in log]

        sample_logs = self.get_sample_tables(subset_successfully_scored_samples, failed_scored_samples)

        logs.update(sample_logs)

        if clear:
            self.clear_logs()

        return logs



from typing import Literal, NamedTuple

from harmony_client.runtime.data import CustomJudgeExample

from adaptive_harmony import StringThread

SYSTEM = """You are an expert evaluator that evaluates completions generated by an AI model on a fixed criterion.
You will be given all elements of an interaction between human an AI model:
    The full context of the conversation so far leading up to the last user turn/question is under the CONTEXT header. It may contain extra contextual information.
    The last user turn/question is under the USER QUESTION header. It may contain extra contextual information.
    The model's completion is under the COMPLETION TO EVALUATE header.
    The evaluation criterion is under the EVALUATION CRITERION section.

In order to to analyze and score a completion, you always run the following steps without exception:
First, you read the CONTEXT, USER QUESTION and COMPLETION TO EVALUATE.
Then, you analyze the COMPLETION TO EVALUATE, and assign it a PASS, FAIL or NA score according to the criterion: FAIL if the completion does not meet the criterion, PASS if it does, and NA if the criterion is not applicable to the example. These are the rules to follow:
- You must always evaluate the COMPLETION TO EVALUATE based solely on the USER QUESTION, and never on an intermediary question that might have been asked in the CONTEXT. The CONTEXT is there for context only.
- Do not make any judgement on text that is in the CONTEXT or USER QUESTION; you are evaluating the COMPLETION TO EVALUATE text only.
- You must not use the original instructions given to the model in the CONTEXT for your judgement. Focus only on the EVALUATION CRITERION, without any other influencing factors.
- You are forbidden to return a score other than PASS, FAIL or NA for each criterion.
- If the criterion is conditional, and is not applicable to the specific USER QUESTION + COMPLETION TO EVALUATE pair, you must score it as NA.
- Return a single score, no matter how many things are evaluated or contemplated in the criterion. A PASS means the completion complied with everything.

Finally, output an explanation for your judgement and the score for the criterion, as exemplified below. The output should be a well-formatted JSON string that conforms to the JSON schema below. Do not output anything else other than the JSON string.

Here is the output JSON schema you must strictly follow, with field descriptions and value types. All fields are required.
{json_schema}

Evaluate only the final COMPLETION TO EVALUATE with regard to the USER QUESTION shown. Do not return any preamble or explanations, Return exactly one JSON string"""


USER = """CONTEXT
{context}

USER QUESTION
{user_question}

COMPLETION TO EVALUATE
{completion}

EVALUATION CRITERION
{criteria}

OUTPUT
"""


class BinaryJudgeShot(NamedTuple):
    """
    Example shot for BinaryJudgeScorer.
    If no criteria is specified, we assume the rest
    of the shot relates to the criteria passed to the BinaryJudgeScorer.
    """

    thread: StringThread
    reasoning: str | None
    score: Literal["PASS", "FAIL", "NA"]
    criteria: str | None = None


def to_binary_judge_shot(example: CustomJudgeExample) -> BinaryJudgeShot:
    return BinaryJudgeShot(
        thread=StringThread([(turn.role, turn.content) for turn in example.input]).assistant(example.output),
        reasoning=example.reasoning,
        score="PASS" if example.pass_ else "FAIL",
    )


DEFAULT_SHOTS = [
    BinaryJudgeShot(
        thread=StringThread(
            [
                (
                    "system",
                    "You are a savy boating conosseur. You know everything there is to know about boats. Your responses to questions must be quick and short, almost dismissive.",
                ),
                ("user", " Hey, how are you doing?"),
                ("assistant", "Doing great!"),
                ("user", "Perfect. Tell me, how many masts does a ketch have?"),
                ("assistant", "A ketch has two masts"),
            ]
        ),
        reasoning="The completion is in English instead of french.",
        score="FAIL",
        criteria="The completion must be in french",
    ),
    BinaryJudgeShot(
        thread=StringThread(
            [
                (
                    "system",
                    "You are a customer service assistant. You can use informal tone if the user is informal with you, but always make sure the customer's problem is fixed.",
                ),
                (
                    "user",
                    "Write a reply to a customer called John Smith that has requested for a refund of our product. Our company policy is to provide refunds up to 30 days after the purchase.",
                ),
                (
                    "assistant",
                    "Dear John Smith,\nThank you for reaching out regarding your request for a refund. We appreciate your business and understand that sometimes products may not meet expectations.\nAs per our company policy, we are happy to provide refunds within 30 days of your purchase. If your request falls within this timeframe and the product is in its original condition, we will process your refund promptly. Please ensure that you return the item to us along with proof of purchase.\nIf you have any questions or need assistance with the return process, feel free to contact us.",
                ),
            ]
        ),
        reasoning="The completion is in general polite to the customer.",
        score="PASS",
        criteria="Completion must be polite, with use of proper language and no slang.",
    ),
    BinaryJudgeShot(
        thread=StringThread(
            [
                (
                    "system",
                    "You are a savy boating conosseur. You know everything there is to know about boats. Your responses to questions must be quick and short, almost dismissive.",
                ),
                ("user", " Hey, how are you doing?"),
                ("assistant", "Doing great!"),
                ("user", "Perfect. Tell me, how many masts does a ketch have?"),
                ("assistant", "A ketch has two masts"),
            ]
        ),
        reasoning="The user did not ask about hunting, criterion is not applicable",
        score="NA",
        criteria="If the user is asking about hunting topics, state that you cannot answer.",
    ),
]
